from ..linalg.sparse import csr_matrix
from .utils import rooted_level_structure, pseudo_peripheral_nodes

__all__ = ["Graph"]

try:
    import networkx as ntx
    import numpy as np

    try:
        adjacency_matrix = ntx.to_scipy_sparse_array
    except Exception:  # pragma: no cover
        adjacency_matrix = ntx.adjacency_matrix

    class Graph(ntx.Graph):
        """
        A subclass of `networkx.Graph`, extending its capabilities.
        See the documentation of `networkx` for the details on how to
        define graphs.

        Note
        ----
        If `networkx` is not installed, the class is `NoneType`, but the
        functionality it implements is still available, you just have to
        manage graph creation by yourself.

        Examples
        --------
        A basic example with `networkx`:

        >>> from sigmaepsilon.math.graph import Graph
        >>> import networkx as nx
        >>> grid = nx.grid_2d_graph(5, 5)  # 5x5 grid
        >>> G = Graph(grid)

        """

        def adjacency_matrix(self, *args, to_csr: bool = False, **kwargs) -> csr_matrix:
            """
            Returns the adjacency matrix of the graph.

            Parameters
            ----------
            to_csr : bool, Optional
                If `True`, the result of networkx.adjacency_matrix is
                returned as a csr_matrix.
            *args : Tuple, Optional
                Forwarded to networkx.adjacency_matrix
            **kwargs, dict, Optional
                Forwarded to networkx.adjacency_matrix

            Returns
            -------
            NumPy array, SciPy array or csr_matrix
                The adjacency representation of the graph.

            Examples
            --------
            >>> from sigmaepsilon.math.graph import Graph
            >>> G = Graph([(1, 1)])
            >>> A = G.adjacency_matrix()
            >>> print(A.todense())
            [[1]]

            """
            adj = adjacency_matrix(self, *args, **kwargs)
            return csr_matrix(adj) if to_csr else adj

        def rooted_level_structure(self, root: int = 0) -> dict[int, np.ndarray]:
            """
            Returns the rooted level structure (RLS) of the graph.

            The call is forwarded to `rooted_level_structure`, go there
            to read about the possible arguments.

            See Also
            --------
            :func:`rooted_level_structure`
            """
            return rooted_level_structure(csr_matrix(adjacency_matrix(self)), root)

        def pseudo_peripheral_nodes(self) -> np.ndarray:
            """
            Returns the indices of nodes that are possible candidates
            for being peripheral nodes of a graph.
            """
            return pseudo_peripheral_nodes(csr_matrix(adjacency_matrix(self)))

except ImportError:  # pragma: no cover
    Graph = None



from enum import Enum
import operator as op
from typing import Callable

from sigmaepsilon.core.kwargtools import getasany

from .function import Function
from .utils import parse_expression, has_operator, valid_operators


__all__ = ["Equality", "InEquality", "Relation"]


class Relations(Enum):
    eq = "="
    gt = ">"
    ge = ">="
    lt = "<"
    le = "<="

    def to_opfunc(self) -> Callable:
        """
        Returns the operator associated with the relation.
        """
        if self == Relations.eq:
            return op.eq
        elif self == Relations.gt:
            return op.gt
        elif self == Relations.ge:
            return op.ge
        elif self == Relations.lt:
            return op.lt
        elif self == Relations.le:
            return op.le
        else:  # pragma: no cover
            raise ValueError(f"Unsupported relation: {self}")

    def to_latex(self) -> str:
        """
        Returns the LaTeX representation of the relation.
        """
        if self == Relations.eq:
            return "="
        elif self == Relations.gt:
            return ">"
        elif self == Relations.ge:
            return r"\geq"
        elif self == Relations.lt:
            return "<"
        elif self == Relations.le:
            return r"\leq"
        else:  # pragma: no cover
            raise ValueError(f"Unsupported relation: {self}")


class Relation(Function):
    """
    Class to express relations. You can use this class to express equalities
    and inequalities. The class is a subclass of :class:`~sigmaepsilon.math.function.function.Function`
    and can be instantiated similatly, with an additional parameter `op` to specify the operator.

    Parameters
    ----------
    op or operator : str or callable
        The operator to be used in the relation. If a string is provided, it should be one of
        =, >, >=, <, <=. If a callable is provided, it should implement a binary operator.
    op_str: {'=', '>', '>=', '<', '<='}, Optional
        If the operator is defined with a callable, you can define the symbol for the operator as
        a string. This is optional and only used in some cases, for instance when generating
        the LaTeX representation of the relation. Also, providing this parameter will help the
        constructor to tetermine the class of the instance when calling the `__new__` method.

    See Also
    --------
    :class:`~sigmaepsilon.math.function.function.Function`
    :class:`~sigmaepsilon.math.function.relation.Equality`
    :class:`~sigmaepsilon.math.function.relation.InEquality`
    """

    __slots__ = ["op", "opfunc", "slack", "op_str"]

    def __new__(cls, *args, op_str: str | None = None, **kwargs):
        # Determine the operator from the arguments
        op = getasany(["op", "operator"], None, **kwargs)

        string_based_input = isinstance(args[0], str) if len(args) == 1 else False
        has_op_in_input = has_operator(args[0]) if string_based_input else False

        if isinstance(op, str):
            if op not in valid_operators:
                raise ValueError(f"Invalid operator: {op}")

            if not op_str:
                op_str = op

            op = Relations(op)

        if not op and has_op_in_input:
            _, operator, _ = parse_expression(args[0])
            op_str = op
            op = Relations(operator)

        if not op and not has_op_in_input:
            if op_str in valid_operators:
                op = Relations(op_str)
            else:
                op = Relations.eq
                op_str = "="

        if op:
            if isinstance(op, Relations):
                op_str = op.value
            elif op_str in valid_operators:
                op = Relations(op_str)

        if cls is not Relation:
            instance = super().__new__(cls)
        elif op == Relations.eq:
            instance = super().__new__(Equality)
        elif op in {Relations.gt, Relations.ge, Relations.lt, Relations.le}:
            instance = super().__new__(InEquality)
        else:
            instance = super().__new__(cls)

        return instance

    def __init__(self, *args, op_str: str | None = None, **kwargs):
        self.op = None
        self.opfunc = None
        self.slack = 0
        self.op_str = op_str

        string_based_input = isinstance(args[0], str) if len(args) == 1 else False
        has_op_in_input = has_operator(args[0]) if string_based_input else False

        op = getasany(["op", "operator"], None, **kwargs)

        if op:
            if isinstance(op, str):
                if op not in valid_operators:
                    raise ValueError(f"Invalid operator: {op}")
                self.op = op = Relations(op)
            elif isinstance(op, Relations):
                self.op = op
            elif isinstance(op, Callable):
                self.opfunc = op
                self.op = None

        if not op and not has_op_in_input:
            if op_str in valid_operators:
                self.op = op = Relations(op_str)
            else:
                self.op = op = Relations.eq

        if not op and has_op_in_input:
            lhs, operator, rhs = parse_expression(args[0])
            rhs = "(" + rhs + ")"
            args = (" - ".join([lhs, rhs]),)
            self.op = op = Relations(operator)

        if op and isinstance(self.op, Relations):
            self.opfunc = self.op.to_opfunc()
            self.op_str = self.op.value

        super().__init__(*args, **kwargs)

    @property
    def operator(self) -> Callable:
        """
        Returns the associated operator.
        """
        return self.op

    def to_latex(self) -> str:
        """
        Returns the LaTeX code of the symbolic expression of the instance.
        Only for symbolic relations.
        """
        expr_str = super().to_latex()
        if self.op in Relations:
            return f"{expr_str} {self.op.to_latex()} 0"
        elif self.op_str:
            op = Relations(self.op_str)
            return f"{expr_str} {op.to_latex()} 0"
        else:  # pragma: no cover
            raise NotImplementedError(
                "LaTeX code generation not available for this kind of function definition."
            )

    def relate(self, *args, **kwargs):
        """
        Relates an input and returns True if it is feasible.
        """
        return self.opfunc(self.f0(*args, **kwargs), 0)


class Equality(Relation):
    """
    Class for equalities, mostly used for expressing
    constraints in optimization problems.

    Examples
    --------
    >>> from sigmaepsilon.math.function import Equality
    >>> import sympy as sy
    >>> variables = ['x1', 'x2', 'x3', 'x4']
    >>> x1, x2, x3, x4 = syms = sy.symbols(variables, positive=True)
    >>> eq1 = Equality(x1 + 2*x3 + x4 - 4, variables=syms)
    >>> eq2 = Equality(x2 + x3 - x4 - 2, variables=syms)

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        assert hasattr(self, "op")
        if self.op:
            assert self.op == Relations.eq


class InEquality(Relation):
    """
    Class for inequalities, mostly used for expressing
    constraints in optimization problems.

    Examples
    --------
    >>> from sigmaepsilon.math.function import InEquality
    >>> gt = InEquality('x + y', op='>')
    >>> ge = InEquality('x + y', op='>=')
    >>> le = InEquality('x + y', op=lambda x, y: x <= y)
    >>> lt = InEquality('x + y', op=lambda x, y: x < y)

    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        assert hasattr(self, "op")
        if self.op:
            assert self.op in {Relations.gt, Relations.ge, Relations.lt, Relations.le}



from typing import Iterable, Callable

import sympy as sy
from sympy import latex
import numpy as np

from sigmaepsilon.deepdict import DeepDict


__all__ = ["gen_Lagrange_1d", "approx_Lagrange_1d"]


def _var_tmpl(i: int):
    return r"\Delta_{}".format(i)


def _var_str(inds, i: int):
    return _var_tmpl(inds[i])


def _diff(xvar, fnc: sy.Expr):
    return fnc.diff(xvar).expand().simplify().factor().simplify()


def gen_Lagrange_1d(
    *_,
    x: Iterable | None = None,
    i: Iterable[int] = None,
    xsym: str | None = None,
    fsym: str | None = None,
    sym: bool = False,
    N: int | None = None,
    lambdify: bool = False,
    out: dict | None = None
) -> dict:
    """
    Generates Lagrange polynomials and their derivatives up to 3rd, for approximation
    in 1d space, based on N input pairs of position and value. Geometrical parameters
    can be numeric or symbolic.

    Parameters
    ----------
    x: Iterable, Optional
        The locations of the data points. If not specified and `sym=False`, a range
        of [-1, 1] is assumed and the locations are generated as `np.linspace(-1, 1, N)`,
        where N is the number of data points. If `sym=True`, the calculation is entirely
        symbolic. Default is None.
    i: Iterable[int]
        If not specified, indices are assumed as [1, ..., N], but this is only relevant
        for symbolic representation, not the calculation itself, which only cares about
        the number of data points, regardless of their actual indices.
    xsym: str, Optional
        Symbol of the variable in the symbolic representation of the generated functions.
        Default is :math:`x`.
    fsym: str, Optional
        Symbol of the function in the symbolic representation of the generated functions.
        Default is 'f'.
    sym: bool, Optional.
        If True, locations of the data points are left in a symbolic state. This requires
        the inversion of a symbolic matrix, which has some reasonable limitations.
        Default is False.
    N: int, Optional
        If neither 'x' nor 'i' is specified, this controls the number of functions to
        generate. Default is None.
    lambdify: bool, Optional
        If True, the functions are turned into `NumPy` functions via `sympy.lambdify`
        and stored in the output for each index with keyword 'fnc'. Default is False.
    out: dict, Optional
        A dictionary to store the values in. Default is None.

    Returns
    -------
    dict
        A dictionary containing the generated functions for the reuested nodes.
        The keys of the dictionary are the indices of the points, the values are
        dictionaries with the following keys and values:

            'symbol' : the `SymPy` symbol of the function

            0 : the function

            1 : the first derivative as a `SymPy` expression

            2 : the second derivative as a `SymPy` expression

            3 : the third derivative as a `SymPy` expression

    Examples
    --------
    >>> from sigmaepsilon.math.approx import gen_Lagrange_1d

    To generate approximation functions for a 2-noded line:

    >>> functions = gen_Lagrange_1d(x=[-1, 1])

    or equivalently

    >>> functions = gen_Lagrange_1d(N=2)

    To generate the same functions in symbolic form:

    >>> functions = gen_Lagrange_1d(i=[1, 2], sym=True)

    Notes
    -----
    Inversion of a heavily symbolic matrix may take quite some time, and is not suggested
    for N > 3. Fixing the locations as constant real numbers symplifies the process and
    makes the solution much faster.
    """
    xsym = xsym if xsym is not None else r"x"
    fsym = fsym if fsym is not None else r"\phi"
    module_data = DeepDict() if not isinstance(out, dict) else out

    xvar = sy.symbols(xsym)
    if not isinstance(N, int):
        assert (x is not None) or (i is not None), "'N', 'x' or 'i' must be provided!"
        N = len(x) if x is not None else len(i)
    inds = list(range(1, N + 1)) if i is None else i

    coeffs = sy.symbols(", ".join(["c_{}".format(i + 1) for i in range(N)]))
    variables = sy.symbols(", ".join([_var_str(inds, i) for i in range(N)]))
    if x is None:
        if not sym:
            x = np.linspace(-1, 1, N)
        else:
            symbols = [xsym + "_{}".format(i + 1) for i in range(N)]
            x = sy.symbols(", ".join(symbols))
    poly = sum([c * xvar**i for i, c in enumerate(coeffs)])

    evals = [poly.subs({xsym: x[i]}) for i in range(N)]
    A = sy.zeros(N, N)
    for i in range(N):
        A[i, :] = sy.Matrix([evals[i].coeff(c) for c in coeffs]).T
    coeffs_new = A.inv() * sy.Matrix(variables)
    subs = {coeffs[i]: coeffs_new[i] for i in range(N)}
    approx = poly.subs(subs).simplify().expand()

    shp = [approx.coeff(v).factor().simplify() for v in variables]

    dshp1 = [_diff(xvar, fnc) for fnc in shp]
    dshp2 = [_diff(xvar, fnc) for fnc in dshp1]
    dshp3 = [_diff(xvar, fnc) for fnc in dshp2]

    for i, ind in enumerate(inds):
        fnc_str = latex(sy.symbols(fsym + "_{}".format(ind)))
        module_data[ind]["symbol"] = fnc_str
        module_data[ind][0] = shp[i]
        module_data[ind][1] = dshp1[i]
        module_data[ind][2] = dshp2[i]
        module_data[ind][3] = dshp3[i]

    if lambdify:
        assert not sym, "If 'lambdify' is True, 'sym' must be False"
        for ind in inds:
            for j in range(4):
                fnc = module_data[ind][j]
                module_data[ind]["fnc"][j] = sy.lambdify(xvar, fnc, "numpy")

    if isinstance(module_data, DeepDict):
        module_data.lock()

    return module_data


def approx_Lagrange_1d(
    points: Iterable, values: Iterable, lambdify: bool = False
) -> Callable:
    """
    Returns a callable that maps from 'source' to 'target' in 1d.

    Parameters
    ----------
    points: Iterable[float]
        The locations of the data points.
    values: Iterable[float]
        The values at the data points.
    lambdify: bool, Optional
        If `True`, the returned function is turned into a `NumPy` function via
        `sympy.lambdify`, otherwise it is left as a `SymPy` expression. Default is `False`.

    Returns
    -------
    Callable
        A vectorized function if 'lambdify' is True, or a SymPy expression otherwise.

    Example
    -------
    >>> from sigmaepsilon.math.approx import approx_Lagrange_1d
    >>> approx = approx_Lagrange_1d([-1, 1], [0, 10], lambdify=True)
    >>> approx(-1), approx(1), approx(0)
    (0, 10, 5)

    A symbolic example:

    >>> import sympy as sy
    >>> L = sy.symbols('L', real=True, positive=True)
    >>> fnc = approx_Lagrange_1d([-1, 1], [0, L])
    >>> str(fnc)
    'L*(x/2 + 1/2)'

    To get the Jacobian of the transformation [0, L] -> [-1, 1]:

    >>> L = sy.symbols('L', real=True, positive=True)
    >>> fnc = approx_Lagrange_1d([0, L], [-1, 1])
    >>> dfnc = fnc.diff('x')
    >>> str(dfnc)
    '2/L'

    Or the other way:

    >>> L = sy.symbols('L', real=True, positive=True)
    >>> fnc = approx_Lagrange_1d([-1, 1], [0, L])
    >>> dfnc = fnc.diff('x')
    >>> str(dfnc)
    'L/2'

    """
    xsym = "x"
    assert len(points) == len(values), "'source' and 'target' must have the same length"
    indices = list(range(len(points)))
    basis = gen_Lagrange_1d(x=points, i=indices, xsym=xsym, lambdify=False)
    shp0 = [basis[i, 0] for i in indices]
    fnc0 = sum([a * f for a, f in zip(values, shp0)])
    if lambdify:
        return sy.lambdify(xsym, fnc0, "numpy")
    else:
        return fnc0



from typing import Callable, Tuple

import numpy as np
from numpy import outer, ndarray
from scipy.special import factorial as fact

from .functions import MLSWeightFunction


def _mls_preproc(
    points: ndarray,
    values: ndarray,
    deg: int,
    dim: int,
    w: MLSWeightFunction,
    b: Callable,
    g: bool = True,
    H: bool = True,
):
    if dim == 1:
        return _mls_preproc_1d(points, values, deg, w, b, g, H)
    elif dim == 2:
        return _mls_preproc_2d(points, values, deg, w, b, g, H)
    elif dim == 3:
        return _mls_preproc_3d(points, values, deg, w, b, g, H)
    else:  # pragma: no cover
        raise NotImplementedError(
            "Only 1, 2 and 3 dimensional pointclouds are supported."
        )


def _mls_preproc_1d(
    points: ndarray,
    values: ndarray,
    deg: int,
    w: MLSWeightFunction,
    b: Callable,
    g: bool = True,
    H: bool = True,
) -> Tuple[ndarray]:
    nData = points.shape[0]
    nDim = 1 if len(points.shape) == 1 else points.shape[1]
    nRec = 1 if len(values.shape) == 1 else values.shape[1]

    k = int(fact(deg + nDim) / fact(deg) / fact(nDim))
    A = np.zeros([k, k])
    B = np.zeros([k, nData])
    Adx, Bdx = None, None
    Adxx, Bdxx = None, None

    if g:
        Adx = np.zeros([k, k])
        Bdx = np.zeros([k, nData])

    if H:
        Adxx = np.zeros([k, k])
        Bdxx = np.zeros([k, nData])

    V = np.zeros([nData, nRec])
    for i in range(nData):
        xi = points[i]
        fi = values[i]
        bi = b(xi)
        wi = w.f(xi)
        Mi = outer(bi, bi)
        A += Mi * wi
        B[:, i] += bi * wi
        if g:
            gwi = w.g(xi)
            Adx += Mi * gwi[0]
            Bdx[:, i] += bi * gwi[0]
        if H:
            Gwi = w.G(xi)
            Adxx += Mi * Gwi[0, 0]
            Bdxx[:, i] += bi * Gwi[0, 0]
        for r in range(nRec):
            V[i, r] = fi[r]

    invA = np.linalg.inv(A)

    return invA, V, B, Adx, Adxx, Bdx, Bdxx


def _mls_preproc_2d(
    points: ndarray,
    values: ndarray,
    deg: int,
    w: MLSWeightFunction,
    b: Callable,
    g: bool = True,
    H: bool = True,
) -> Tuple[ndarray]:
    nData = points.shape[0]
    nDim = points.shape[1]
    nRec = values.shape[1]

    k = int(fact(deg + nDim) / fact(deg) / fact(nDim))
    A = np.zeros([k, k])
    B = np.zeros([k, nData])
    Adx, Ady, Bdx, Bdy = 4 * (None,)
    Adxx, Adyy, Adxy, Bdxx, Bdyy, Bdxy = 6 * (None,)

    if g:
        Adx = np.zeros([k, k])
        Ady = np.zeros([k, k])
        Bdx = np.zeros([k, nData])
        Bdy = np.zeros([k, nData])

    if H:
        Adxx = np.zeros([k, k])
        Adyy = np.zeros([k, k])
        Adxy = np.zeros([k, k])
        Bdxx = np.zeros([k, nData])
        Bdyy = np.zeros([k, nData])
        Bdxy = np.zeros([k, nData])

    V = np.zeros([nData, nRec])
    for i in range(nData):
        xi = points[i]
        fi = values[i]
        bi = b(xi)
        wi = w.f(xi)
        Mi = outer(bi, bi)
        A += Mi * wi
        B[:, i] += bi * wi
        if g:
            gwi = w.g(xi)
            Adx += Mi * gwi[0]
            Ady += Mi * gwi[1]
            Bdx[:, i] += bi * gwi[0]
            Bdy[:, i] += bi * gwi[1]
        if H:
            Gwi = w.G(xi)
            Adxx += Mi * Gwi[0, 0]
            Adyy += Mi * Gwi[1, 1]
            Adxy += Mi * Gwi[0, 1]
            Bdxx[:, i] += bi * Gwi[0, 0]
            Bdyy[:, i] += bi * Gwi[1, 1]
            Bdxy[:, i] += bi * Gwi[0, 1]
        for r in range(nRec):
            V[i, r] = fi[r]

    invA = np.linalg.inv(A)

    return invA, V, B, Adx, Ady, Adxx, Adyy, Adxy, Bdx, Bdy, Bdxx, Bdyy, Bdxy


def _mls_preproc_3d(
    points: ndarray,
    values: ndarray,
    deg: int,
    w: MLSWeightFunction,
    b: Callable,
    g: bool = True,
    H: bool = True,
) -> Tuple[ndarray]:
    nData = points.shape[0]
    nDim = points.shape[1]
    nRec = values.shape[1]

    k = int(fact(deg + nDim) / fact(deg) / fact(nDim))
    A = np.zeros([k, k])
    B = np.zeros([k, nData])
    Adx, Ady, Adz, Bdx, Bdy, Bdz = 6 * (None,)
    Adxx, Adyy, Adzz, Adxy, Adxz, Adyz, Bdxx, Bdyy, Bdzz, Bdxy, Bdxz, Bdyz = 12 * (
        None,
    )

    if g:
        Adx = np.zeros([k, k])
        Ady = np.zeros([k, k])
        Adz = np.zeros([k, k])
        Bdx = np.zeros([k, nData])
        Bdy = np.zeros([k, nData])
        Bdz = np.zeros([k, nData])

    if H:
        Adxx = np.zeros([k, k])
        Adyy = np.zeros([k, k])
        Adzz = np.zeros([k, k])
        Adxy = np.zeros([k, k])
        Adxz = np.zeros([k, k])
        Adyz = np.zeros([k, k])
        Bdxx = np.zeros([k, nData])
        Bdyy = np.zeros([k, nData])
        Bdzz = np.zeros([k, nData])
        Bdxy = np.zeros([k, nData])
        Bdxz = np.zeros([k, nData])
        Bdyz = np.zeros([k, nData])

    V = np.zeros([nData, nRec])
    for i in range(nData):
        xi = points[i]
        fi = values[i]
        bi = b(xi)
        wi = w.f(xi)
        Mi = outer(bi, bi)
        A += Mi * wi
        B[:, i] += bi * wi
        if g:
            gwi = w.g(xi)
            Adx += Mi * gwi[0]
            Ady += Mi * gwi[1]
            Adz += Mi * gwi[2]
            Bdx[:, i] += bi * gwi[0]
            Bdy[:, i] += bi * gwi[1]
            Bdz[:, i] += bi * gwi[2]
        if H:
            Gwi = w.G(xi)
            Adxx += Mi * Gwi[0, 0]
            Adyy += Mi * Gwi[1, 1]
            Adzz += Mi * Gwi[2, 2]
            Adxy += Mi * Gwi[0, 1]
            Adxz += Mi * Gwi[0, 2]
            Adyz += Mi * Gwi[1, 2]
            Bdxx[:, i] += bi * Gwi[0, 0]
            Bdyy[:, i] += bi * Gwi[1, 1]
            Bdzz[:, i] += bi * Gwi[2, 2]
            Bdxy[:, i] += bi * Gwi[0, 1]
            Bdxz[:, i] += bi * Gwi[0, 2]
            Bdyz[:, i] += bi * Gwi[1, 2]
        for r in range(nRec):
            V[i, r] = fi[r]

    invA = np.linalg.inv(A)

    return (
        invA,
        V,
        B,
        Adx,
        Ady,
        Adz,
        Adxx,
        Adyy,
        Adzz,
        Adxy,
        Adxz,
        Adyz,
        Bdx,
        Bdy,
        Bdz,
        Bdxx,
        Bdyy,
        Bdzz,
        Bdxy,
        Bdxz,
        Bdyz,
    )



from typing import Tuple, Any, Iterable
from numbers import Number

import numpy as np
from numpy.linalg import norm
from numpy import ndarray

from ..function import Function


class MLSWeightFunction(Function):
    """
    Base class for weight functions for the moving least squares method.
    """

    def __init__(
        self,
        *,
        core: int | Iterable[Number] | ndarray | None = None,
        supportdomain: Iterable[Number] | None = None,
        sd: Iterable[Number] | None = None,
        **kwargs,
    ):
        super().__init__(**kwargs)

        dim = None
        if not isinstance(core, ndarray):
            if isinstance(core, Iterable):
                core = np.array(core)
                dim = core.shape[0]
            elif isinstance(core, int):
                dim = 1
                core = np.zeros([dim])
        else:
            dim = core.shape[0]

        if dim is not None:
            self.dimension = dim

        if not any([sd is None, supportdomain is None]):
            raise ValueError(
                "`supportdomain` and `sd` cannot be both specified at the same time."
            )
        sd = sd if sd is not None else supportdomain

        self._core = core
        self._supportdomain = sd

    @property
    def core(self) -> ndarray | Iterable[Number] | None:
        return self._core

    @core.setter
    def core(self, val: ndarray | Iterable[Number] | Number | None):
        if not isinstance(val, ndarray):
            if isinstance(val, Iterable):
                val = np.array(val)
            elif isinstance(val, Number):
                val = np.array([val])
            else:
                raise ValueError(
                    f"Expected a NumPy ndarray, or an Iterable, got {type(val)}"
                )
        self._core = val
        return

    @property
    def supportdomain(self) -> ndarray | Iterable[Number] | None:
        return self._supportdomain

    @supportdomain.setter
    def supportdomain(self, val: ndarray | Iterable[Number] | None):
        if not isinstance(val, ndarray):
            if isinstance(val, Iterable):
                val = np.array(val)
            else:
                raise ValueError(
                    f"Expected a NumPy ndarray, or an Iterable, got {type(val)}"
                )
        self._supportdomain = val
        return

    def preproc_evaluation(self, x: Iterable[Number]):
        if not isinstance(x, (ndarray, float)):
            if isinstance(x, Iterable):
                x = np.array(x)
            else:
                raise ValueError(
                    f"Expected a NumPy ndarray, or an Iterable, got {type(x)}"
                )

    def value(self, x: Iterable[Number]) -> float:
        """
        Evaluates the function.
        """
        raise NotImplementedError


def isMLSWeightFunction(f: Any) -> bool:
    """
    Returns `True` if the argument is a valid weight function for the
    moving least squares method.
    """
    c1 = isinstance(f, MLSWeightFunction)
    c2 = MLSWeightFunction in list(type(f).__bases__)
    return any([c1, c2])


class ConstantWeightFunction(MLSWeightFunction):
    """
    A constant weight function for the moving least squares method.
    """

    def __init__(self, *, value: Number = 1.0, **kwargs):
        super().__init__(**kwargs)
        self._value = value
        return

    def value(self, x: Iterable[Number]) -> float:
        if self.supportdomain is None:
            return self._value

        d = np.subtract(self.core, x)
        r = np.abs(d / np.array(self.supportdomain))

        if any(r > 1):
            return 0

        return self._value

    def gradient(self, x: Iterable[Number]) -> ndarray:
        return np.zeros(self.dimension, dtype=float)

    def Hessian(self, x: Iterable[Number]) -> ndarray:
        return np.zeros((self.dimension, self.dimension), dtype=float)


class SingularWeightFunction(MLSWeightFunction):
    """
    A singular weight function for the moving least squares method.
    """

    def __init__(self, *, eps: Number = 1e-5, **kwargs):
        super().__init__(**kwargs)
        self.eps = eps
        return

    def value(self, x: Iterable[Number]):
        self.preproc_evaluation(x)
        return 1 / (norm(np.subtract(self.core, x)) ** 2 + self.eps**2)

    def gradient(self, x: Iterable[Number]) -> ndarray:
        return np.zeros(self.dimension, dtype=float)

    def Hessian(self, x: Iterable[Number]) -> ndarray:
        return np.zeros((self.dimension, self.dimension), dtype=float)


class CubicWeightFunction(MLSWeightFunction):
    """
    A cubic weight function for the moving least squares method.

    Example
    -------
    >>> from sigmaepsilon.math.approx import CubicWeightFunction
    >>> w = CubicWeightFunction(core=[0.0, 0.0], sd=[0.5, 0.5])
    >>> w([0.0, 0.0])
    0.4444444444444444

    """

    def evaluate(self, x: Iterable[Number]) -> Tuple[float, ndarray, ndarray]:
        if self.dimension == 1:
            return self._evaluate_1d(x)
        elif self.dimension == 2:
            return self._evaluate_2d(x)

        raise NotImplementedError

    def _evaluate_1d(self, x: Iterable[Number]) -> Tuple[float, ndarray, ndarray]:
        d = np.subtract(self.core, x)
        difX = d[0]
        dmX = self.supportdomain[0]
        rX = abs(difX) / dmX

        if abs(difX) < 1e-12:
            drdX = 0
        else:
            drdX = (difX / abs(difX)) / dmX

        if rX <= 0.5:
            wX = 2 / 3 - 4 * rX**2 + 4 * rX**3
            dwXdX = (-8 * rX + 12 * rX**2) * drdX
            dwXdXX = (-8 + 24 * rX) * drdX * drdX
        elif rX > 0.5 and rX <= 1:
            wX = 4 / 3 - 4 * rX + 4 * rX**2 - (4 / 3) * rX**3
            dwXdX = (-4 + 8 * rX - 4 * rX**2) * drdX
            dwXdXX = (8 - 8 * rX) * drdX * drdX
        else:
            wX = 0
            dwXdX = 0
            dwXdXX = 0

        val = wX
        grad = np.array([dwXdX])
        Hessian = np.array([[dwXdXX]])

        return val, grad, Hessian

    def _evaluate_2d(self, x: Iterable[Number]) -> Tuple[float, ndarray, ndarray]:
        d = np.subtract(self.core, x)
        difX = d[0]
        difY = d[1]
        dmX = self.supportdomain[0]
        dmY = self.supportdomain[1]
        rX = abs(difX) / dmX
        rY = abs(difY) / dmY

        if abs(difX) < 1e-12:
            drdX = 0
        else:
            drdX = (difX / abs(difX)) / dmX

        if abs(difY) < 1e-12:
            drdY = 0
        else:
            drdY = (difY / abs(difY)) / dmY

        if rX <= 0.5:
            wX = 2 / 3 - 4 * rX**2 + 4 * rX**3
            dwXdX = (-8 * rX + 12 * rX**2) * drdX
            dwXdXX = (-8 + 24 * rX) * drdX * drdX
        elif rX > 0.5 and rX <= 1:
            wX = 4 / 3 - 4 * rX + 4 * rX**2 - (4 / 3) * rX**3
            dwXdX = (-4 + 8 * rX - 4 * rX**2) * drdX
            dwXdXX = (8 - 8 * rX) * drdX * drdX
        else:
            wX = 0
            dwXdX = 0
            dwXdXX = 0

        if rY <= 0.5:
            wY = 2 / 3 - 4 * rY**2 + 4 * rY**3
            dwYdY = (-8 * rY + 12 * rY**2) * drdY
            dwYdYY = (-8 + 24 * rY) * drdY * drdY
        elif rY > 0.5 and rY <= 1:
            wY = 4 / 3 - 4 * rY + 4 * rY**2 - (4 / 3) * rY**3
            dwYdY = (-4 + 8 * rY - 4 * rY**2) * drdY
            dwYdYY = (8 - 8 * rY) * drdY * drdY
        else:
            wY = 0
            dwYdY = 0
            dwYdYY = 0

        val = wX * wY
        grad = np.array([wY * dwXdX, wX * dwYdY])
        Hessian = np.array([[wY * dwXdXX, dwXdX * dwYdY], [dwXdX * dwYdY, wX * dwYdYY]])

        return val, grad, Hessian

    def value(self, x: Iterable[Number]) -> float:
        self.preproc_evaluation(x)
        res, _, _ = self.evaluate(x)
        return res

    def gradient(self, x: Iterable[Number]) -> ndarray:
        self.preproc_evaluation(x)
        _, res, _ = self.evaluate(x)
        return res

    def Hessian(self, x: Iterable[Number]) -> ndarray:
        self.preproc_evaluation(x)
        _, _, res = self.evaluate(x)
        return res



from typing import Any, Type


def _new_and_init_(cls: Type, *args: Any, **kwargs: Any) -> Any:
    """
    Create a new instance of a class and initialize it.
    """
    obj = cls.__new__(cls, *args, **kwargs)
    obj.__init__(*args, **kwargs)
    return obj



from copy import deepcopy

import numpy as np

from sigmaepsilon.core.testing import SigmaEpsilonTestCase

from .meta import TensorLike


class LinalgTestCase(SigmaEpsilonTestCase):
    def assertTensorDualityRules(self, t: TensorLike) -> None:
        # test if the dual of the dual is self
        self.assertTrue(np.allclose(t.dual().dual().show(), t.show()))

    def assertTensorTranspositionRules(self, t: TensorLike) -> None:
        # test if A.T.T == A
        self.assertTrue(np.allclose(t.show(), t.T.T.show()))

        # test if (A + B).T == A.T + B.T
        t_ = deepcopy(t)
        t_.array *= 2
        t__ = t + t_
        self.assertTrue(np.allclose(t__.T.show(), t.T.show() + t_.T.show()))

        # test if (A + B).T == A.T + B.T
        t__ = t - t_
        self.assertTrue(np.allclose(t__.T.show(), t.T.show() - t_.T.show()))

    def assertSymmetricTensor(self, t: TensorLike) -> None:
        # test if A - A.T == 0
        self.assertTrue(np.allclose(np.zeros_like(t.array), t.show() - t.T.show()))



from copy import deepcopy as dcopy

import numpy as np
from numpy import ndarray

from sigmaepsilon.core.alphabet import latinrange

from .frame import ReferenceFrame as Frame
from .abstract import AbstractTensor
from .tr import _tr_tensors2, _tr_tensors4x3
from .exceptions import TensorShapeMismatchError
from .utils import transpose_axes
from .logical import is_hermitian

__all__ = [
    "Tensor",
    "Tensor2",
    "Tensor4",
    "Tensor2x3",
    "Tensor4x3",
]


class Tensor(AbstractTensor):
    """
    A class to handle tensors.

    Parameters
    ----------
    args: tuple, Optional
        Positional arguments forwarded to `numpy.ndarray`.
    frame: numpy.ndarray, Optional
        The reference frame the vector is represented by its coordinates.
    kwargs: dict, Optional
        Keyword arguments forwarded to `numpy.ndarray`.

    Examples
    --------
    Import the necessary classes:

    >>> from sigmaepsilon.math.linalg import Tensor, ReferenceFrame
    >>> from numpy.random import rand

    Create a Tensor of order 6 in a frame with random components

    >>> frame = ReferenceFrame(dim=3)
    >>> array = rand(3, 3, 3, 3, 3, 3)
    >>> A = Tensor(array, frame=frame)

    Get the tensor in the dual frame:

    >>> A_dual = A.dual()

    Create an other tensor, in this case a 5th-order one, and calculate their
    generalized dot product, which is a 9th-order tensor:

    >>> from sigmaepsilon.math.linalg import dot
    >>> array = rand(3, 3, 3, 3, 3)
    >>> B = Tensor(array, frame=frame)
    >>> C = dot(A, B, axes=[0, 0])
    >>> assert C.rank == (A.rank + B.rank - 2)

    See Also
    --------
    :class:`~sigmaepsilon.math.linalg.vector.Vector`
    :class:`~sigmaepsilon.math.linalg.frame.ReferenceFrame`
    """

    _frame_cls_ = Frame
    _einsum_params_ = {}

    @classmethod
    def _verify_input(cls, arr: ndarray, *_, **kwargs) -> bool:
        return is_hermitian(arr)

    @classmethod
    def _from_any_input(cls, *args, **kwargs) -> AbstractTensor:
        if cls._verify_input(*args, **kwargs):
            return cls(*args, **kwargs)
        else:
            if not Tensor._verify_input(*args, **kwargs):
                raise ValueError("Invalid input to Tensor class.")
            else:
                return Tensor(*args, **kwargs)

    def dual(self) -> "Tensor2":
        """
        Returns the tensor described in the dual (or reciprocal) frame.
        """
        a = self.transform_components(self.frame.Gram())
        return self.__class__(a, frame=self.frame.dual())

    def transform_components(self, Q: ndarray) -> ndarray:
        """
        Returns the components of the tensor transformed by the matrix Q.
        """
        r = self.rank
        arr = self.array
        args = [Q for _ in range(r)]
        if r not in self.__class__._einsum_params_:
            target = latinrange(r, start=ord("a"))
            source = latinrange(r, start=ord("a") + r)
            terms = [t + s for s, t in zip(source, target)]
            command = ",".join(terms) + "," + "".join(source)
            einsum_path = np.einsum_path(command, *args, arr, optimize="greedy")[0]
            self.__class__._einsum_params_[r] = (command, einsum_path)
        else:
            command, einsum_path = self.__class__._einsum_params_[r]
        return np.einsum(command, *args, arr, optimize=einsum_path)

    def show(self, target: Frame = None, *, dcm: ndarray = None) -> ndarray:
        """
        Returns the components in a target frame. If the target is
        `None`, the components are returned in the ambient frame.

        The transformation can also be specified with a proper DCM matrix.

        Parameters
        ----------
        target: numpy.ndarray, Optional
            Target frame.
        dcm: numpy.ndarray, Optional
            The DCM matrix of the transformation.

        Returns
        -------
        numpy.ndarray
            The components of the tensor in a specified frame, or
            the ambient frame, depending on the arguments.
        """
        if not isinstance(dcm, ndarray):
            if target is None:
                target = Frame(dim=self.frame.axes.shape[-1])
            dcm = self.frame.dcm(target=target)
        return self.transform_components(dcm)

    def orient(self, *args, **kwargs) -> "Tensor":
        """
        Orients the vector inplace. All arguments are forwarded to
        `orient_new`.

        Returns
        -------
        Vector
            The same vector the function is called upon.

        See Also
        --------
        :func:`orient_new`
        """
        fcls: Frame = self.__class__._frame_cls_
        dcm = fcls.eye(dim=self.frame.shape[-1]).orient_new(*args, **kwargs).dcm()
        self.array = self.transform_components(transpose_axes(dcm))
        return self

    def orient_new(self, *args, **kwargs) -> "Tensor":
        """
        Returns a transformed version of the instance.

        Returns
        -------
        Vector
            A new vector.

        See Also
        --------
        :func:`orient`
        """
        fcls: Frame = self.__class__._frame_cls_
        dcm = fcls.eye(dim=self.frame.shape[-1]).orient_new(*args, **kwargs).dcm()
        array = self.transform_components(dcm.T)
        return self.__class__(array, frame=self.frame)

    def copy(self, deep: bool = False, name: str = None) -> "Tensor":
        """
        Returns a shallow or deep copy of this object, depending of the
        argument `deepcopy` (default is False).
        """
        if deep:
            return self.__class__(dcopy(self.array), name=name)
        else:
            return self.__class__(self.array, name=name)

    def deepcopy(self, name: str = None) -> "Tensor":
        """
        Returns a deep copy of the frame.
        """
        return self.copy(deep=True, name=name)


class Tensor2(Tensor):
    """
    A class to handle second-order tensors. Some operations have dedicated implementations
    that provide higher performence utilizing implicit parallelization. Examples
    for tensors of this class include the metric tensor, or the stress and strain tensors
    of elasticity.

    See also
    --------
    :class:`~sigmaepsilon.math.linalg.tensor.Tensor2x3`
    """

    _rank_ = 2

    @classmethod
    def _verify_input(cls, arr: ndarray, *_, bulk: bool = False, **kwargs) -> bool:
        if bulk:
            return len(arr.shape) == 3 and arr.shape[-1] == arr.shape[-2]
        else:
            return len(arr.shape) == 2 and arr.shape[-1] == arr.shape[-2]

    def transform_components(self, Q: ndarray) -> ndarray:
        return _tr_tensors2(self.array, Q)


class Tensor2x3(Tensor2):
    """
    Dedicated class for second-order tensors, with 3 indices per axis.
    Since the shape of the tensor is known, instances are able to automatically detect
    if the provided components resemble a single item or a collection.
    """

    def __init__(self, *args, **kwargs):
        if len(args) > 0 and isinstance(args[0], ndarray):
            arr = args[0]
            shape = arr.shape
            if shape[-2:] == (3, 3):
                if len(shape) >= 3:
                    is_bulk = kwargs.get("bulk", True)
                    if not is_bulk:
                        raise ValueError("Incorrect input!")
                    kwargs["bulk"] = is_bulk
                else:
                    if not len(shape) == 2:
                        raise TensorShapeMismatchError("Invalid shape!")
                    is_bulk = kwargs.get("bulk", False)
                    if is_bulk:
                        raise ValueError("Incorrect input!")
            else:
                raise TensorShapeMismatchError("Invalid shape!")

        super().__init__(*args, **kwargs)

    @classmethod
    def _verify_input(cls, arr: ndarray, *_, bulk: bool = False, **kwargs) -> bool:
        if bulk:
            return len(arr.shape) >= 3 and arr.shape[-2:] == (3, 3)
        else:
            return len(arr.shape) == 2 and arr.shape[-2:] == (3, 3)


class Tensor4(Tensor):
    """
    A class to handle fourth-order tensors. Some operations have dedicated implementations
    that provide higher performence utilizing implicit parallelization. Examples of this class
    include the piezo-optical tensor, the elasto-optical tensor, the flexoelectric tensor or the
    elasticity tensor.

    See also
    --------
    :class:`~sigmaepsilon.math.linalg.tensor.Tensor4x3`
    """

    _rank_ = 4

    @classmethod
    def _verify_input(cls, arr: ndarray, *_, bulk: bool = False, **kwargs) -> bool:
        shape = arr.shape
        is_hermitian = (shape[-1],) * 4 == shape[-4:]
        if bulk:
            return len(shape) == 5 and is_hermitian
        else:
            return len(shape) == 4 and is_hermitian

    def transform_components(self, dcm: ndarray) -> ndarray:
        """
        Returns the components of the transformed numerical tensor, based on
        the provided direction cosine matrix.
        """
        return _tr_tensors4x3(self._array, dcm)


class Tensor4x3(Tensor4):
    """
    Dedicated class for fourth order tensors, with 3 indices per axis.
    Since the shape of the tensor is known, instances are able to automatically detect
    if the provided components resemble a single item or a collection.
    """

    def __init__(self, *args, **kwargs):
        if len(args) > 0 and isinstance(args[0], ndarray):
            arr = args[0]
            shape = arr.shape
            if shape[-4:] == (3, 3, 3, 3):
                if len(shape) >= 5:
                    is_bulk = kwargs.get("bulk", True)
                    if not is_bulk:
                        raise ValueError("Incorrect input!")
                    kwargs["bulk"] = is_bulk
                else:
                    if not len(shape) == 4:
                        raise TensorShapeMismatchError("Invalid shape!")
                    is_bulk = kwargs.get("bulk", False)
                    if is_bulk:
                        raise ValueError("Incorrect input!")
            else:
                raise TensorShapeMismatchError("Invalid shape!")

        super().__init__(*args, **kwargs)

    @classmethod
    def _verify_input(cls, arr: ndarray, *_, bulk: bool = False, **kwargs) -> bool:
        is_hermitian = arr.shape[-4:] == (3, 3, 3, 3)
        if bulk:
            return len(arr.shape) >= 5 and is_hermitian
        else:
            return len(arr.shape) == 4 and is_hermitian



from typing import Callable, Iterable

import numpy as np
from numpy import ndarray

from .functions import isMLSWeightFunction, ConstantWeightFunction
from .ls_poly import _get_polynomial
from .ls_approx import _mls_approx
from .ls_preproc import _mls_preproc

__all__ = ["moving_least_squares", "least_squares", "weighted_least_squares"]


def moving_least_squares(
    points: ndarray | Iterable,
    values: ndarray | Iterable,
    *,
    w: Callable | None = None,
    **kwargs,
) -> Callable:
    """
    Moving least squares approximation. The usage is the same as for the
    :func:`~sigmaepsilon.math.approx.ls.weighted_least_squares` function.
    """
    dim = 1 if len(points.shape) == 1 else points.shape[1]

    if not isMLSWeightFunction(w):
        w = ConstantWeightFunction(dim=dim)

    def inner(x):
        if not isinstance(x, np.ndarray):
            if isinstance(x, Iterable):
                x = np.array(x)
            else:
                if dim > 1:
                    raise TypeError(
                        f"Invalid input type. Expected a numpy array or an iterable, got {type(x)}"
                    )
        w.core = x
        f = weighted_least_squares(points, values, w=w, **kwargs)
        return f(x)

    return inner


def least_squares(
    points: ndarray | Iterable,
    values: ndarray | Iterable,
    *,
    deg: int = 1,
    order: int = 2,
) -> Callable:
    """
    Given :math:`N` points located at :math:`\mathbf{x}_i` in :math:`\mathbb{R}^d`
    where :math:`i \in [1 \dots N]`. The returned fit function approximates the given
    values :math:`f_i` at :math:`\mathbf{x}_i` in the least-squares sence with the
    error functional

    .. math::
        \sum_{i} \\biggr[ || f \left( \mathbf{x}_i \\right) - f_i || \\biggr] ^2

    where :math:`f` is taken from :math:`\Pi_{m}^d`, the space of polynomials of
    total degree :math:`m` in :math:`d` spatial dimensions.

    Parameters
    ----------
    points: Iterable
        [[X11, X12, ..., X1d], ..., [Xn1, Xn2, ..., Xnd]]
    values: Iterable
        [[f11, f12, ..., f1r], ..., [fn1, fn2, ..., fnr]]
    deg: int, Optional
        The degree of the fit function. Default is 1.
    order: int, Optional.
        The order of the approximation. Default is 2.

    Returns
    -------
    Callable
        Fit function r(x) -> f(x), fdx(x), fdy(x), fdxx(x), fdyy(x), fdxy(x)
        fi([X1, X2, ..., Xd]) = [fi1, fi2,..., fir]

    Note
    ----
    The resulting fit function can have an approximation or regression behaviour,
    depending on the dataset and the degree of the polynomial.
    """
    return weighted_least_squares(points, values, deg=deg, order=order)


def weighted_least_squares(
    points: ndarray | Iterable,
    values: ndarray | Iterable,
    *,
    deg: int = 1,
    order: int = 2,
    w: Callable | None = None,
) -> Callable:
    """
    Returns a Callable that can be used to approximate over datasets.

    Parameters
    ----------
    points: Iterable
        [[X11, X12, ..., X1d], ..., [Xn1, Xn2, ..., Xnd]]
    values: Iterable
        [[f11, f12, ..., f1r], ..., [fn1, fn2, ..., fnr]]
    deg: int, Optional
        The degree of the fit function. Default is 1.
    w: MLSWeightFunction, Optional
        A proper weight function. Default is a `ConstantWeightFunction`.
    order: int, Optional.
        The order of the approximation. Default is 2.

    Returns
    -------
    Callable
        Fit function r(x) -> f(x), fdx(x), fdy(x), fdxx(x), fdyy(x), fdxy(x)
        fi([X1, X2, ..., Xd]) = [fi1, fi2,..., fir]

    Note
    ----
    The resulting fit function can have an approximation or regression behaviour,
    depending on the dataset and the degree of the polynomial.
    """
    if not isinstance(points, ndarray):
        points = np.array(points)

    if not isinstance(values, ndarray):
        values = np.array(values)

    assert isinstance(points, ndarray)
    assert isinstance(values, ndarray)

    if not points.shape[0] == values.shape[0]:
        raise ValueError(
            (
                f"The number of points ({len(points)}) does "
                "not match the number of values ({len(values)})."
            )
        )

    if len(values.shape) == 1:
        values = values.reshape(len(values), 1)

    dim = 1 if len(points.shape) == 1 else points.shape[1]

    if isMLSWeightFunction(w):
        assert dim == w.dimension
    else:
        w = ConstantWeightFunction(dim=dim)

    grad = True if order > 0 else False
    hess = True if order > 1 else False

    if hess:
        grad = True
        assert hasattr(w, "Hessian")

    if grad:
        assert hasattr(w, "gradient")

    if dim == 1:
        return _wls_1d(points, values, deg, dim, w, grad, hess)
    elif dim == 2:
        return _wls_2d(points, values, deg, dim, w, grad, hess)
    elif dim == 3:
        return _wls_3d(points, values, deg, dim, w, grad, hess)
    else:
        raise NotImplementedError(
            "Only 1, 2 and 3 dimensional pointclouds are supported"
        )


def _wls_1d(points, values, deg, dim, w, grad, hess):
    b, bdx, bdxx = _get_polynomial(deg, dim)

    (invA, V, B, Adx, Adxx, Bdx, Bdxx) = _mls_preproc(
        points, values, deg, dim, w, b, grad, hess
    )

    def inner(x):
        return _mls_approx(
            invA,
            B,
            b(x),
            V,
            Adx,
            Adxx,
            Bdx,
            Bdxx,
            bdx(x),
            bdxx(x),
            g=grad,
            H=hess,
            dim=dim,
        )

    return inner


def _wls_2d(points, values, deg, dim, w, grad, hess):
    b, bdx, bdy, bdxx, bdyy, bdxy = _get_polynomial(deg, dim)

    (invA, V, B, Adx, Ady, Adxx, Adyy, Adxy, Bdx, Bdy, Bdxx, Bdyy, Bdxy) = _mls_preproc(
        points, values, deg, dim, w, b, grad, hess
    )

    def inner(x):
        return _mls_approx(
            invA,
            B,
            b(x),
            V,
            Adx,
            Ady,
            Adxx,
            Adyy,
            Adxy,
            Bdx,
            Bdy,
            Bdxx,
            Bdyy,
            Bdxy,
            bdx(x),
            bdy(x),
            bdxx(x),
            bdyy(x),
            bdxy(x),
            g=grad,
            H=hess,
            dim=dim,
        )

    return inner


def _wls_3d(points, values, deg, dim, w, grad, hess):
    b, bdx, bdy, bdz, bdxx, bdyy, bdzz, bdxy, bdxz, bdyz = _get_polynomial(deg, dim)

    (
        invA,
        V,
        B,
        Adx,
        Ady,
        Adz,
        Adxx,
        Adyy,
        Adzz,
        Adxy,
        Adxz,
        Adyz,
        Bdx,
        Bdy,
        Bdz,
        Bdxx,
        Bdyy,
        Bdzz,
        Bdxy,
        Bdxz,
        Bdyz,
    ) = _mls_preproc(points, values, deg, dim, w, b, grad, hess)

    def inner(x):
        return _mls_approx(
            invA,
            B,
            b(x),
            V,
            Adx,
            Ady,
            Adz,
            Adxx,
            Adyy,
            Adzz,
            Adxy,
            Adxz,
            Adyz,
            Bdx,
            Bdy,
            Bdz,
            Bdxx,
            Bdyy,
            Bdzz,
            Bdxy,
            Bdxz,
            Bdyz,
            bdx(x),
            bdy(x),
            bdz(x),
            bdxx(x),
            bdyy(x),
            bdzz(x),
            bdxy(x),
            bdxz(x),
            bdyz(x),
            g=grad,
            H=hess,
            dim=dim,
        )

    return inner



from numbers import Number

from numpy import ndarray
import numpy as np
from numba import njit, prange

from ..knn import k_nearest_neighbours
from .. import atleast2d

__all__ = ["MLSApproximator"]


@njit(parallel=True, cache=True)
def _approximate_2d(values: ndarray, neighbours: ndarray, factors: ndarray) -> ndarray:
    nT, nN = neighbours.shape
    nD = values.shape[-1]
    res = np.zeros((nT, nD), dtype=values.dtype)
    for i in prange(nT):
        for j in prange(nD):
            for k in range(nN):
                res[i, j] += values[neighbours[i, k], j] * factors[i, k]
    return res


@njit(parallel=True, cache=True)
def _approximate_nd(
    values: ndarray, neighbours: ndarray, factors: ndarray, out: ndarray
) -> ndarray:
    nT, nN = neighbours.shape
    for i in prange(nT):
        for j in range(nN):
            out[i] += values[neighbours[i, j]] * factors[i, j]


class MLSApproximator:
    """
    Object oriented, high performance implementation of a specific version of the
    moving least squares method. This implementation is less flexible than the others,
    but performes well for extremely large datasets too. If you want to experiment
    with the hyperparameters of the MLS as a method, it is suggested to use the other
    solutions offered by the library.

    Parameters
    ----------
    X_S: ndarray
        The source points.
    Y_S: ndarray, Optional
        The source data.
    knn_backend: {"scipy", "sklearn"}, Optional
        The backend to use for the KNN calculation. If None, the default backend
        of the library is used. Default is None.
    k: int, Optional
        The number of neighbours to consider. Default is None.
    max_distance: Number, Optional
        The maximum distance to consider for the neighbours. Default is None.

    Notes
    -----
    1) There is a KNN calculation involved in the process, which is done using either
    scipy or sklearn.
    2) The target points can be provided when calling the :func:`fit` method to precalculate
    the neighbours and factors. If not provided, the neighbours and factors are calculated
    when the :func:`approximate` method is called.
    3) We use Numba to speed up the approximation, hence the first call might be slower.

    Examples
    --------
    >>> import numpy as np
    >>> from sigmaepsilon.math.approx import MLSApproximator
    >>>
    >>> # prepare the source points
    >>> nx, ny, nz = (10, 10, 10)
    >>> x = np.linspace(0, 1, nx)
    >>> y = np.linspace(0, 1, ny)
    >>> z = np.linspace(0, 1, nz)
    >>> xv, yv, zv = np.meshgrid(x, y, z)
    >>> source_points = np.stack([xv.flatten(), yv.flatten(), zv.flatten()], axis=-1)
    >>>
    >>> # prepare the source values
    >>> source_values = np.ones((nx*ny*nz, 3))  # could be anything with a shape of (nx*ny*nz, ...)
    >>>
    >>> # instantiate the approximator
    >>> approximator = MLSApproximator(source_points, source_values, knn_backend="scipy")
    >>>
    >>> # approximate to the target points
    >>> target_points = source_points[:10]  # could be anything with a shape of (..., source_points.shape[-1])
    >>> target_values = approximator.approximate(target_points)
    >>>
    >>> # check the results
    >>> assert np.allclose(target_values, np.ones_like(target_values))

    """

    __slots__ = ["X_S", "Y_S", "_neighbours", "_factors", "_knn_config"]

    def __init__(
        self,
        X_S: ndarray,
        Y_S: ndarray,
        knn_backend: str | None = None,
        k: int | None = None,
        max_distance: Number | None = None,
    ) -> None:
        self.X_S = X_S
        self.Y_S = Y_S
        self._neighbours = None
        self._factors = None
        self._knn_config = dict()

        if isinstance(knn_backend, str):
            self._knn_config["knn_backend"] = knn_backend
        elif knn_backend is not None:  # pragma: no cover
            raise ValueError(
                f"Expected a string for parameter 'knn_backend', got {type(knn_backend)} instead."
            )

        if isinstance(k, int):
            self._knn_config["k"] = k
        elif k is not None:  # pragma: no cover
            raise ValueError(
                f"Expected an integer for parameter 'k', got {type(k)} instead."
            )

        if isinstance(max_distance, Number):
            self._knn_config["max_distance"] = float(max_distance)
        elif max_distance is not None:  # pragma: no cover
            raise ValueError(
                f"Expected a number for parameter 'max_distance', got {type(max_distance)} instead."
            )

    @property
    def neighbours(self) -> ndarray | None:
        """
        Returns the neighbours of the target points. If the neighbours are not
        calculated yet, it returns None.
        """
        return self._neighbours

    @neighbours.setter
    def neighbours(self, val: ndarray) -> None:
        """
        Sets the neighbours of the target points.
        """
        self._neighbours = val

    @property
    def factors(self) -> ndarray | None:
        """
        Returns the factors of the target points. If the factors are not
        calculated yet, it returns None.
        """
        return self._factors

    @factors.setter
    def factors(self, val: ndarray) -> None:
        """
        Sets the factors of the target points.
        """
        self._factors = val

    def _calc_factors_and_neighbours(self, X_S: ndarray, X_T: ndarray) -> None:
        self.neighbours = self._get_neighbours(X_S, X_T, **self._knn_config)
        self.factors = np.ones_like(self.neighbours) / self.neighbours.shape[-1]

    @staticmethod
    def _get_neighbours(
        X_S: ndarray,
        X_T,
        *,
        k: int = 4,
        max_distance: float | None = None,
        knn_backend: str = "scipy",
    ) -> ndarray:
        k = 4 if not k else k
        return k_nearest_neighbours(
            X_S, X_T, k=k, max_distance=max_distance, backend=knn_backend
        )

    def approximate(self, X_T: ndarray) -> ndarray:
        """
        Estimates the value of the function at the given points.

        Parameters
        ----------
        X_T: ndarray
            The target points.
        """
        X_S = atleast2d(self.X_S, back=True)
        X_T = atleast2d(X_T, back=True)
        self._calc_factors_and_neighbours(X_S, X_T)
        neighbours = self.neighbours
        factors = self.factors

        data = self.Y_S
        if len(self.Y_S.shape) == 1:
            data = atleast2d(data, back=True)

        if len(data.shape) == 2:
            res = _approximate_2d(data, neighbours, factors)
        else:
            res = np.zeros((len(neighbours),) + data.shape[1:], dtype=data.dtype)
            _approximate_nd(data, neighbours, factors, out=res)

        return np.squeeze(res)



import re

valid_operators = ["<=", ">=", "<", ">", "="]


def parse_expression(expression: str) -> tuple[str, str, str]:
    """
    Takes an expression and returns a tuple with the left side, operator and right
    side of the expression.

    Example
    -------
    >>> from sigmaepsilon.math.function.utils import parse_expression
    >>> parse_expression("x <= 2")
    ('x', '<=', '2')

    """
    # Regular expression to match the pattern (left side, operator, right side)
    pattern = r"(.+?)\s*(<=|>=|<|>|=)\s*(.+)"

    # Use re.match to extract parts of the expression
    match = re.match(pattern, expression)

    if match:
        left_side = match.group(1).strip()
        operator = match.group(2)
        right_side = match.group(3).strip()

        for op in valid_operators:
            if op in left_side:
                raise ValueError(f"Invalid expression: '{expression}'")

            if op in right_side:
                raise ValueError(f"Invalid expression: '{expression}'")

        if operator not in valid_operators:
            raise ValueError(f"Invalid expression: '{expression}'")

        return left_side, operator, right_side
    else:
        raise ValueError(f"Invalid expression: '{expression}'")


def has_operator(expression: str) -> bool:
    """
    Returns True if the expression contains a valid operator.

    Example
    -------
    >>> from sigmaepsilon.math.function.utils import has_operator
    >>> has_operator("x <= 2")
    True

    """
    for op in valid_operators:
        if op in expression:
            return True

    return False



from sympy.parsing.sympy_parser import parse_expr
from sympy import lambdify, derive_by_array, symbols, Expr, Symbol
from sympy.core.numbers import One
from typing import Iterable
import sympy as sy


def generate_symbols(
    template: str, indices: Iterable[int], **assumptions
) -> list[Symbol]:
    """
    Generates a list of symbols.
    """
    result = list(
        sy.symbols(" ".join([template.format(i) for i in indices]), **assumptions)
    )
    return result


def decode(
    *_,
    expr: Expr | None = None,
    str_expr: str | None = None,
    variables: Iterable | None = None,
    **__,
) -> tuple[Expr, Iterable[Symbol]]:
    """
    Takes an expression as either a string or a `SymPy` expression and returns
    the expression and the variables in the expression.
    """
    if str_expr is not None:
        expr = parse_expr(str_expr, evaluate=False)

    if not variables:
        variables = []
        variables = tuple(expr.free_symbols)
    else:
        if not all([isinstance(v, (sy.Expr, sy.Symbol)) for v in variables]):
            variables = list(symbols(variables))

        expr = substitute(expr, variables, variables, as_string=True)

    return expr, variables


def symbolize(*args, simplify: bool = True, **kwargs) -> dict:
    expr, variables = decode(*args, **kwargs)
    if simplify:
        expr = expr.simplify()
    f0 = lambdify([variables], expr, "numpy")
    g = derive_by_array(expr, variables)
    f1 = lambdify([variables], g, "numpy")
    G = derive_by_array(g, variables)
    f2 = lambdify([variables], G, "numpy")
    return {
        "value": f0,
        "gradient": f1,
        "Hessian": f2,
        "d": len(variables),
        "variables": variables,
        "expr": expr,
    }


def substitute(
    expr: Expr,
    values: Iterable,
    variables: Iterable[str | Symbol] | None = None,
    as_string: bool = False,
) -> Expr:
    """
    Substitutes the variables in the expression with the corresponding values.

    Parameters
    ----------
    expr : sympy.Expr
        The expression to substitute the values in.
    values : list
        The values to substitute the variables with.
    variables : list, optional
        The variables to substitute. If not provided, the function will attempt to
        extract the variables from the expression.
    as_string : bool, optional
        If True, the variables will be substituted as strings. Default is `False`.
        This might be important if let say you want to instantiate a Function with
        the string 'x + y' and you want the expression to have variables that you
        prepared with `sympy.symbols`. In that case, two SymPy variables with the same
        name will be treated as different variables. Substituting the variables as strings
        will help you avoid this issue.
    """
    if variables is None:
        variables = tuple(expr.free_symbols)

    if not as_string:
        return expr.subs([(v, val) for v, val in zip(variables, values)])
    else:
        return expr.subs([(str(v), val) for v, val in zip(variables, values)])


def coefficients(
    expr: Expr, variables: Iterable | None = None, normalize: bool = False
) -> dict:
    """
    Returns the coefficients of the expression.

    Parameters
    ----------
    expr : sympy.Expr
        The expression to extract the coefficients from.
    variables : list, optional
        The variables to extract the coefficients from. If not provided, the function
        will attempt to extract the variables from the expression.
    normalize : bool, optional
        If True, the constant term (if present) in the expression will be normalized.
        See the examples for more details.

    Examples
    --------
    The following example illustrated the effect ot the `normalize` parameter.

    >>> from sigmaepsilon.math.function import Function
    >>> from sigmaepsilon.math.function.symutils import coefficients
    >>> g = Function("3*x + 4*y - 2", variables=["x", "y", "z"])
    >>> coefficients(g.expr)  # doctest: +SKIP
    {x: 3, y: 4, 2: -1}

    >>> coefficients(g.expr, normalize=True)  # doctest: +SKIP
    {x: 3, y: 4, 1: -2}

    We can see, that the constant term in the expression is handled differently.
    In the first case, the constant term is viewed as the number 2 with coefficient -1,
    while in the second case, the constant term is viewed as the number 1 with coefficient -2.
    """
    if variables is None:
        variables = tuple(expr.free_symbols)

    d = dict({x: 0 for x in variables})
    d.update(expr.as_coefficients_dict())

    if not normalize:
        return d
    else:
        res = dict()
        for key, value in d.items():
            if len(key.free_symbols) == 0:
                res[One()] = value * key
            else:
                res[key] = value
        return res



from typing import Iterable, Sequence

import numpy as np
from numpy import ndarray
from sympy import Symbol
from scipy.optimize import linprog, OptimizeResult

from ..function import Function, InEquality, Equality
from ..function.relation import Relations, Relation
from ..mathtypes import BoundsLike

__all__ = ["LinearProgrammingProblem"]


class LinearProgrammingProblem:
    """
    A class to solve linear programming problems [1]_. It uses the :func:`scipy.optimize.linprog`
    function as a solver [2]_, which eventually calls into the HIGHS solver [3]_.
    
    To define the objective and the constraints,
    you can use the `Function`, `Relation`, `Equality` and `InEquality` 
    classes. These are able to understand SymPy expressions and strings, giving you
    the flexibility in defining the problem.

    Parameters
    ----------
    obj : Function
        The objective function.
    constraints : Sequence[Relation]
        The constraints.
    variables : Sequence[Symbol], Optional
        The variables of the problem. The variables must be provided for symbolic functions
        to guarantee that the order of the variables is consistent with the symbolic functions
        that make up the problem.
    bounds : BoundsLike, Optional
        See the `bounds` parameter in `scipy.optimize.linprog` for more details.
    integrality : Iterable[int] | int, Optional
        See the `integrality` parameter in `scipy.optimize.linprog` for more details.
        
    See Also
    --------
    :class:`~sigmaepsilon.math.function.function.Function`
    :class:`~sigmaepsilon.math.function.relation.Relation`
    :class:`~sigmaepsilon.math.function.relation.Equality`
    :class:`~sigmaepsilon.math.function.relation.InEquality`

    Examples
    --------
    The following example solves a problem with a unique solution.

    .. math::

        \\begin{eqnarray}
            & minimize&  \\quad  3 x_1 + x_2 + 9 x_3 + x_4  \\\\
            & subject \\, to& & \\nonumber\\\\
            & & x_1 + 2 x_3 + x_4 \\,=\\, 4, \\\\
            & & x_2 + x_3 - x_4 \\,=\\, 2, \\\\
            & & x_i \\,\\geq\\, \\, 0, \\qquad i=1, \\ldots, 4.
        \\end{eqnarray}

    >>> from sigmaepsilon.math.optimize import LinearProgrammingProblem as LPP
    >>> from sigmaepsilon.math.function import Function, Relation
    >>> import sympy as sy
    >>> 
    >>> x1, x2, x3, x4 = variables = sy.symbols('x1:5')
    >>> obj = Function(3*x1 + 9*x3 + x2 + x4, variables=variables)
    >>> eq1 = Relation(x1 + 2*x3 + x4 - 4, variables=variables)
    >>> eq2 = Relation(x2 + x3 - x4 - 2, variables=variables)
    >>>
    >>> bounds = [(0, None), (0, None), (0, None), (0, None)]
    >>> problem = LPP(obj, [eq1, eq2], variables=variables, bounds=bounds)
    >>> problem.solve().x
    array([0., 6., 0., 4.])
    
    In this example, bounds could have been specified as `bounds=(0, None)` as
    well, since all variables have the same bound.
    
    Now let see how to solve the following mixed integer linear programming problem.
    
    .. math::

        \\begin{eqnarray}
            & minimize&  \\quad  3 x_2 + 2 x_3  \\\\
            & subject \\, to& & \\nonumber\\\\
            & & 2 x_1 + 2 x_2 - 4 x_3 \\,=\\, 5, \\\\
            & & x_i \\,\\geq\\, \\, 0, \\qquad i=1, \\ldots, 4. \\\\
            & & x_1, x_3 \\,\\in\\, \\mathbb{Z}.
        \\end{eqnarray}
    
    >>> variables = x1, x2, x3 = sy.symbols(["x1", "x2", "x3"])
    >>> f = Function(3 * x2 + 2 * x3, variables=variables)
    >>> eq = Relation(2 * x1 + 2 * x2 - 4 * x3 - 5, op="=", variables=variables)
    >>> bounds = (0, None)
    >>> integrality = [1, 0, 1]
    >>> problem = LPP(f, [eq], variables=variables, bounds=bounds, integrality=integrality)
    
    These integrality constraints can also be specified using assumptions on the
    symbolic variables.
    
    >>> x1, x3 = sy.symbols(["x1", "x3"], integer=True)
    >>> x2 = sy.symbols("x2")
    >>> variables = x1, x2, x3
    >>> f = Function(3 * x2 + 2 * x3, variables=variables)
    >>> eq = Relation("2 * x1 + 2 * x2 - 4 * x3 = 5", variables=variables)
    >>> problem = LPP(f, [eq], variables=variables, bounds=(0, None))
    
    References
    ----------
    .. [1] https://en.wikipedia.org/wiki/Linear_programming
    .. [2] https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.linprog.html
    .. [3] https://highs.dev/
    """

    __slots__ = [
        "obj",
        "constraints",
        "_variables",
        "bounds",
        "integrality",
    ]

    def __init__(
        self,
        obj: Function,
        constraints: Sequence[Relation],
        *,
        variables: Sequence[Symbol] | None = None,
        bounds: BoundsLike = (0, None),
        integrality: Iterable[int] | int = None,
    ):
        super().__init__()
        self.obj = None
        self.constraints = []
        self._variables = []
        self.bounds = bounds
        self.integrality = integrality

        _variables = []

        if variables is None:
            raise ValueError("The variables must be provided for symbolic problems!")

        if isinstance(obj, Function):
            if not obj.is_symbolic:
                raise ValueError("The objective function must be symbolic!")

            self.obj = obj
            _variables += self.obj.variables
        else:
            raise TypeError("The objective function must be an intance of `Function`!")

        for constraint in constraints:
            if isinstance(constraint, Relation):
                if not constraint.is_symbolic:
                    raise ValueError("The constraints must be symbolic!")

                self.constraints.append(constraint)
                _variables += constraint.variables
            else:
                raise TypeError("The constraints must be instances of `Relation`!")

        _variables = set(_variables)

        if variables is not None:
            if not all([isinstance(v, Symbol) for v in variables]):
                raise TypeError("All variables must be instances of `Symbol`!")

            if not all([v in _variables for v in variables]):
                raise ValueError("Inconsistent variables provided!")

            self._variables = variables

    @property
    def variables(self) -> Iterable[Symbol]:
        """
        Returns the variables of the problem.
        """
        return self._variables

    def _to_scipy(self, *, maximize: bool = False) -> tuple[ndarray, dict]:
        """
        Returns values for the parameters `A_ub`, `b_ub`, `A_eq`, `b_eq`, `bounds` and `integrality`
        for the `scipy.optimize.linprog` function.
        """
        n_eq = len(list(filter(lambda c: isinstance(c, Equality), self.constraints)))
        n_ieq = len(list(filter(lambda c: isinstance(c, InEquality), self.constraints)))
        n_x = len(self.variables)

        coeffs = self.obj.linear_coefficients(normalize=True)
        c = np.array([coeffs[x_] for x_ in self.variables]).astype(float)
        if maximize:
            c *= -1

        A_ub, b_ub, A_eq, b_eq = 4 * [None]

        x_zero = np.zeros((n_x,))
        SMALLNUM = np.nextafter(0, 1)

        if n_eq > 0:
            A_eq = np.zeros((n_eq, n_x))
            b_eq = np.zeros((n_eq,))
            equalities: Iterable[Equality] = filter(
                lambda c: isinstance(c, Equality), self.constraints
            )
            for i, eq in enumerate(equalities):
                coeffs = eq.linear_coefficients(normalize=True)
                A_eq[i] = [coeffs[x_] for x_ in self.variables]
                b_eq[i] = -eq(x_zero)

        if n_ieq > 0:
            A_ub = np.zeros((n_ieq, n_x))
            b_ub = np.zeros((n_ieq,))
            inequalities: Iterable[InEquality] = filter(
                lambda c: isinstance(c, InEquality), self.constraints
            )
            for i, ieq in enumerate(inequalities):
                coeffs = ieq.linear_coefficients(normalize=True)
                A_ub[i] = [coeffs[x_] for x_ in self.variables]
                b_ub[i] = -ieq(x_zero)

                if ieq.op == Relations.ge:
                    A_ub[i] *= -1
                    b_ub[i] *= -1
                elif ieq.op == Relations.gt:
                    A_ub[i] *= -1
                    b_ub[i] *= -1
                    b_ub[i] -= SMALLNUM
                elif ieq.op == Relations.lt:
                    b_ub[i] -= SMALLNUM

        integrality = self.integrality
        if integrality is None:
            is_int = [v.is_integer for v in self.variables]
            if any(is_int):
                integrality = np.array(is_int, dtype=bool).astype(int)

        kwargs = dict(
            A_ub=A_ub,
            b_ub=b_ub,
            A_eq=A_eq,
            b_eq=b_eq,
            bounds=self.bounds,
            integrality=integrality,
        )

        return c, kwargs

    def solve(
        self, *, maximize: bool = False, method: str = "highs", **kwargs
    ) -> OptimizeResult:
        """
        Solves the linear programming problem using `scipy.optimize.linprog`
        and returns an instance of `scipy.optimize.OptimizeResult`.

        Parameters
        ----------
        maximize : bool, Optional
            If `True`, the problem is a maximization problem.
        method : str, Optional
            The solver to use. The default is "highs". For more options see the
            `method` parameter in `scipy.optimize.linprog`.
        **kwargs
            Additional keyword arguments to pass to `scipy.optimize.linprog`.

        Returns
        -------
        :class:`scipy.optimize.OptimizeResult`
            The result of the optimization.
        """
        c, _kwargs = self._to_scipy(maximize=maximize)
        _kwargs.update(kwargs)
        _kwargs["method"] = method
        res = linprog(c, **_kwargs)
        if res.success:
            res.fun = self.obj(res.x)
        return res



import numpy as np


def _get_polynomial(deg: int, dim: int):
    if deg == 1:
        if dim == 1:
            return _get_poly_1_1()
        elif dim == 2:
            return _get_poly_1_2()
        elif dim == 3:
            return _get_poly_1_3()
        else:  # pragma: no cover
            raise NotImplementedError
    elif deg == 2:
        if dim == 1:
            return _get_poly_2_1()
        elif dim == 2:
            return _get_poly_2_2()
        elif dim == 3:
            return _get_poly_2_3()
        else:  # pragma: no cover
            raise NotImplementedError
    else:  # pragma: no cover
        raise NotImplementedError


def _get_poly_1_1():
    def b(x):
        return np.array([1, x])

    def bdx(x):
        return np.array([0, 1])

    def bdxx(x):
        return np.array([0, 0])

    return b, bdx, bdxx


def _get_poly_1_2():
    def b(x):
        return np.array([1, x[0], x[1]])

    def bdx(x):
        return np.array([0, 1, 0])

    def bdy(x):
        return np.array([0, 0, 1])

    def bdxx(x):
        return np.array([0, 0, 0])

    def bdyy(x):
        return np.array([0, 0, 0])

    def bdxy(x):
        return np.array([0, 0, 0])

    return b, bdx, bdy, bdxx, bdyy, bdxy


def _get_poly_1_3():
    def b(x):
        return np.array([1, x[0], x[1], x[2]])

    def bdx(x):
        return np.array([0, 1, 0, 0])

    def bdy(x):
        return np.array([0, 0, 1, 0])

    def bdz(x):
        return np.array([0, 0, 0, 1])

    def bdxx(x):
        return np.array([0, 0, 0, 0])

    def bdyy(x):
        return np.array([0, 0, 0, 0])

    def bdzz(x):
        return np.array([0, 0, 0, 0])

    def bdxy(x):
        return np.array([0, 0, 0, 0])

    def bdxz(x):
        return np.array([0, 0, 0, 0])

    def bdyz(x):
        return np.array([0, 0, 0, 0])

    return b, bdx, bdy, bdz, bdxx, bdyy, bdzz, bdxy, bdxz, bdyz


def _get_poly_2_1():
    def b(x):
        return np.array([1, x, x**2])

    def bdx(x):
        return np.array([0, 1, 2 * x])

    def bdxx(x):
        return np.array([0, 0, 2])

    return b, bdx, bdxx


def _get_poly_2_2():
    def b(x):
        return np.array([1, x[0], x[1], x[0] ** 2, x[1] ** 2, x[0] * x[1]])

    def bdx(x):
        return np.array([0, 1, 0, 2 * x[0], 0, x[1]])

    def bdy(x):
        return np.array([0, 0, 1, 0, 2 * x[1], x[0]])

    def bdxx(x):
        return np.array([0, 0, 0, 2, 0, 0])

    def bdyy(x):
        return np.array([0, 0, 0, 0, 2, 0])

    def bdxy(x):
        return np.array([0, 0, 0, 0, 0, 1])

    return b, bdx, bdy, bdxx, bdyy, bdxy


def _get_poly_2_3():
    def b(x):
        return np.array(
            [
                1,
                x[0],
                x[1],
                x[2],
                x[0] ** 2,
                x[1] ** 2,
                x[2] ** 2,
                x[0] * x[1],
                x[0] * x[2],
                x[1] * x[2],
            ]
        )

    def bdx(x):
        return np.array([0, 1, 0, 0, 2 * x[0], 0, 0, x[1], x[2], 0])

    def bdy(x):
        return np.array([0, 0, 1, 0, 0, 2 * x[1], 0, x[0], 0, x[2]])

    def bdz(x):
        return np.array([0, 0, 0, 1, 0, 0, 2 * x[2], 0, x[0], x[1]])

    def bdxx(x):
        return np.array([0, 0, 0, 0, 2, 0, 0, 0, 0, 0])

    def bdyy(x):
        return np.array([0, 0, 0, 0, 0, 2, 0, 0, 0, 0])

    def bdzz(x):
        return np.array([0, 0, 0, 0, 0, 0, 2, 0, 0, 0])

    def bdxy(x):
        return np.array([0, 0, 0, 0, 0, 0, 0, 1, 0, 0])

    def bdxz(x):
        return np.array([0, 0, 0, 0, 0, 0, 0, 0, 1, 0])

    def bdyz(x):
        return np.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 1])

    return b, bdx, bdy, bdz, bdxx, bdyy, bdzz, bdxy, bdxz, bdyz



from copy import deepcopy as dcopy

from numpy import ndarray
import numbers

from .utils import show_vector
from .frame import ReferenceFrame as Frame
from .abstract import AbstractTensor
from .meta import FrameLike


__all__ = ["Vector"]


class Vector(AbstractTensor):
    """
    Extends `NumPy`'s ``ndarray`` class to handle arrays with associated
    reference frames. The class also provides a mechanism to transform
    vectors between different frames. Use it like if it was a ``numpy.ndarray``
    instance.

    All parameters are identical to those of ``numpy.ndarray``, except that
    this class allows to specify an embedding frame.

    Parameters
    ----------
    args: tuple, Optional
        Positional arguments forwarded to `numpy.ndarray`.
    frame: FrameLike, Optional
        The reference frame the vector is represented by its coordinates.
    kwargs: dict, Optional
        Keyword arguments forwarded to `numpy.ndarray`.

    Examples
    --------
    Import the necessary classes:

    >>> import numpy as np
    >>> from sigmaepsilon.math.linalg import Vector, ReferenceFrame

    Create a default frame in 3d space, and create 2 others, each
    being rotated with 30 degrees around the third axis.

    >>> A = ReferenceFrame(dim=3)
    >>> B = A.orient_new('Body', [0, 0, 30*np.pi/180], 'XYZ')
    >>> C = B.orient_new('Body', [0, 0, 30*np.pi/180], 'XYZ')

    To create a vector in a frame:

    >>> vA = Vector([1.0, 1.0, 0.0], frame=A)

    To create a vector with a relative transformation to another one:

    >>> vB = vA.orient_new('Body', [0, 0, -30*np.pi/180], 'XYZ')

    Use the `array` property to get the componets of a `Vector`:

    >>> vB.array
    Array([1.3660254, 0.3660254, 0.       ])

    If you want to obtain the components of a vector in a specific
    target frame C, do this:

    >>> vB.show(C)
    array([ 1., -1.,  0.])

    The reason why the result is represented now as 'array' insted of 'Array'
    as in the previous case is that the Vector class is an array container. When
    you type `vB.array`, what is returned is a wrapped object, an instance of `Array`,
    which is also a class of this library. When you say `vB.show(C)`, a NumPy array
    is returned. Since the `Array` class is a direct subclass of NumPy's `ndarray` class,
    it doesn't really matter and the only difference is the capital first letter.

    To create a vector in a target frame C, knowing the components in a
    source frame A:

    >>> vC = Vector(vA.show(C), frame=C)

    See Also
    --------
    :class:`~sigmaepsilon.math.linalg.tensor.Tensor`
    :class:`~sigmaepsilon.math.linalg.frame.ReferenceFrame`
    """

    _frame_cls_ = Frame
    _HANDLED_TYPES_ = (numbers.Number,)

    def __init__(
        self,
        *args,
        frame: FrameLike | None = None,
        **kwargs,
    ):
        super().__init__(*args, frame=frame, **kwargs)

    @classmethod
    def _verify_input(cls, arr: ndarray, *_, **kwargs) -> bool:
        """
        Ought to verify if an array input is acceptable for the current class.
        If not a general Tensor class is returned upon calling the creator.
        """
        return True

    @property
    def rank(self) -> int:
        """
        Returns the tensor rank (or order).
        """
        return 1

    def dual(self) -> "Vector":
        """
        Returns the vector described in the dual (or reciprocal) frame.
        """
        # NOTE Strictly this should be self.frame.Gram().T @ self.array,
        # but since the Gram matrix is symmetric, it's cheaper like this
        a = self.frame.Gram() @ self.array
        return self.__class__(a, frame=self.frame.dual())

    def show(self, target: Frame = None, *, dcm: ndarray = None) -> ndarray:
        """
        Returns the components in a target frame. If the target is
        `None`, the components are returned in the ambient frame.

        The transformation can also be specified with a proper DCM matrix.

        Parameters
        ----------
        target: numpy.ndarray, Optional
            Target frame.
        dcm: numpy.ndarray, Optional
            The DCM matrix of the transformation.

        Returns
        -------
        numpy.ndarray
            The components of the vector in a specified frame, or
            the ambient frame, depending on the arguments.
        """
        if not isinstance(dcm, ndarray):
            if target is None:
                target = self._frame_cls_(dim=self._array.shape[-1])
            dcm = self.frame.dcm(target=target)
        return show_vector(dcm, self.array)  # dcm @ arr

    def orient(self, *args, dcm: ndarray = None, **kwargs) -> "Vector":
        """
        Orients the vector inplace. If the transformation is not specified by 'dcm',
        all arguments are forwarded to `orient_new`.

        Parameters
        ----------
        dcm: numpy.ndarray, Optional
            The DCM matrix of the transformation.

        Returns
        -------
        Vector
            The same vector the function is called upon.

        See Also
        --------
        :func:`orient_new`
        """
        if not isinstance(dcm, ndarray):
            fcls = self.__class__._frame_cls_
            dcm = fcls.eye(dim=len(self)).orient_new(*args, **kwargs).dcm()
            # self.array = dcm.T @ self._array
            self.array = show_vector(dcm.T, self.array)
            # self.array = np.linalg.inv(dcm) @ self._array
            # FIXME check this
        else:
            self.array = show_vector(dcm.T, self.array)
            # self.array = dcm.T @ self._array
            # FIXME check if inversion is necessary here
            # inversion might be necessary here because it is uncertain if the
            # dcm matrix was fabricated properly.
            # self.array = np.linalg.inv(dcm) @ self._array
        return self

    def orient_new(self, *args, **kwargs) -> "Vector":
        """
        Returns a transformed version of the instance.

        Returns
        -------
        Vector
            A new vector.

        See Also
        --------
        :func:`orient`
        """
        fcls = self.__class__._frame_cls_
        dcm = fcls.eye(dim=len(self)).orient_new(*args, **kwargs).dcm()
        array = dcm.T @ self._array
        # FIXME check if inversion is necessary or not
        # array = np.linalg.inv(dcm) @ self._array
        return Vector(array, frame=self.frame)

    def copy(self, deep: bool = False, name: str = None) -> "Vector":
        """
        Returns a shallow or deep copy of this object, depending of the
        argument `deepcopy` (default is False).
        """
        if deep:
            return self.__class__(dcopy(self.array), name=name)
        else:
            return self.__class__(self.array, name=name)

    def deepcopy(self, name: str = None) -> "Vector":
        """
        Returns a deep copy of the frame.
        """
        return self.copy(deep=True, name=name)



from typing import Iterable

import numpy as np
from numpy import ndarray
from numba import njit, prange

__cache = True


def _tr_tensors2(arr: ndarray, Q: ndarray) -> ndarray:
    shape_arr = arr.shape
    shape_Q = Q.shape
    n_shape_arr = len(shape_arr)
    n_shape_Q = len(shape_Q)
    if n_shape_arr == 2 and n_shape_Q == 2:
        return Q @ arr @ Q.T
    elif n_shape_arr == 3 and n_shape_Q == 2:
        return _tr_tensors2_single(arr, Q)
    elif n_shape_arr >= 4 and n_shape_Q == 2:
        nX = np.prod(shape_arr[1:-2], dtype=int)
        arr = arr.reshape(shape_arr[0], nX, shape_arr[-2], shape_arr[-1])
        res = _tr_tensors2_single2(arr, Q)
        res = res.reshape(shape_arr)
        return res
    elif n_shape_arr == 3 and n_shape_Q == 3:
        return _tr_tensors2_multi(arr, Q)
    elif n_shape_arr == 4 and n_shape_Q == 3:
        return _tr_tensors2_multi2(arr, Q)
    elif n_shape_arr > 4 and n_shape_Q == 3:
        nX = np.prod(shape_arr[1:-2], dtype=int)
        arr = arr.reshape(shape_arr[0], nX, shape_arr[-2], shape_arr[-1])
        res = _tr_tensors2_multi2(arr, Q)
        res = res.reshape(shape_arr)
        return res
    else:
        raise NotImplementedError


@njit(nogil=True, parallel=True, cache=__cache)
def _tr_tensors2_single(arr: ndarray, Q: ndarray) -> ndarray:
    nE = arr.shape[0]
    res = np.zeros_like(arr)
    Q_T = Q.T
    for iE in prange(nE):
        res[iE, :, :] = Q @ arr[iE] @ Q_T
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def _tr_tensors2_single2(arr: ndarray, Q: ndarray) -> ndarray:
    nE, nX = arr.shape[:2]
    res = np.zeros_like(arr)
    Q_T = Q.T
    for iE in prange(nE):
        for iX in prange(nX):
            res[iE, iX, :, :] = Q @ arr[iE, iX] @ Q_T
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def _tr_tensors2_multi(arr: ndarray, Q: ndarray) -> ndarray:
    nE = arr.shape[0]
    res = np.zeros_like(arr)
    for iE in prange(nE):
        res[iE, :, :] = Q[iE] @ arr[iE] @ Q[iE].T
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def _tr_tensors2_multi2(arr: ndarray, Q: ndarray) -> ndarray:
    nE, nX = arr.shape[:2]
    res = np.zeros_like(arr)
    for iE in prange(nE):
        Qi = Q[iE]
        QiT = Qi.T
        for iX in prange(nX):
            res[iE, iX, :, :] = Qi @ arr[iE, iX] @ QiT
    return res


def _tr_tensors4x3(arr: ndarray, Q: ndarray) -> ndarray:
    shape_arr = arr.shape
    shape_Q = Q.shape
    n_shape_arr = len(shape_arr)
    n_shape_Q = len(shape_Q)
    if n_shape_arr == 4 and n_shape_Q == 2:
        return _tr_3333(arr, Q)
    elif n_shape_arr == 5 and n_shape_Q == 2:
        return _tr_3333_multi3(arr, Q)
    elif n_shape_arr == 5 and n_shape_Q == 3:
        return _tr_3333_multi(arr, Q)
    elif n_shape_arr == 6 and n_shape_Q == 3:
        return _tr_3333_multi2(arr, Q)
    elif n_shape_arr > 6 and n_shape_Q == 3:
        nX = np.prod(shape_arr[1:-4], dtype=int)
        arr = arr.reshape(
            shape_arr[0], nX, shape_arr[-4], shape_arr[-3], shape_arr[-2], shape_arr[-1]
        )
        res = _tr_3333_multi2(arr, Q)
        res = res.reshape(shape_arr)
        return res
    else:
        raise NotImplementedError


def _tr_3333_sym(array: Iterable, Q: ndarray) -> ndarray:
    res = np.zeros((3, 3, 3, 3), dtype=array.dtype)
    for p in prange(3):
        for q in prange(3):
            for r in prange(3):
                for s in prange(3):
                    for i in prange(3):
                        for j in prange(3):
                            for k in prange(3):
                                for m in prange(3):
                                    res[i, j, k, m] += (
                                        Q[i, p]
                                        * Q[j, q]
                                        * Q[k, r]
                                        * Q[m, s]
                                        * array[p, q, r, s]
                                    )
    return res


@njit(nogil=True, parallel=False, fastmath=True, cache=__cache)
def _tr_3333(array: ndarray, Q: ndarray) -> ndarray:
    res = np.zeros((3, 3, 3, 3), dtype=array.dtype)
    for p in prange(3):
        for q in prange(3):
            for r in prange(3):
                for s in prange(3):
                    for i in prange(3):
                        for j in prange(3):
                            for k in prange(3):
                                for m in prange(3):
                                    res[i, j, k, m] += (
                                        Q[i, p]
                                        * Q[j, q]
                                        * Q[k, r]
                                        * Q[m, s]
                                        * array[p, q, r, s]
                                    )
    return res


@njit(nogil=True, parallel=False, fastmath=True, cache=__cache)
def _tr_3333_out(array: ndarray, Q: ndarray, out: ndarray) -> ndarray:
    for p in prange(3):
        for q in prange(3):
            for r in prange(3):
                for s in prange(3):
                    for i in prange(3):
                        for j in prange(3):
                            for k in prange(3):
                                for m in prange(3):
                                    out[i, j, k, m] += (
                                        Q[i, p]
                                        * Q[j, q]
                                        * Q[k, r]
                                        * Q[m, s]
                                        * array[p, q, r, s]
                                    )


@njit(nogil=True, parallel=True, cache=__cache)
def _tr_3333_multi(arr: ndarray, Q: ndarray) -> ndarray:
    nE = arr.shape[0]
    res = np.zeros_like(arr)
    for iE in prange(nE):
        _tr_3333_out(arr[iE, :, :, :, :], Q[iE], res[iE, :, :, :, :])
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def _tr_3333_multi2(arr: ndarray, Q: ndarray) -> ndarray:
    nE, nX = arr.shape[:2]
    res = np.zeros_like(arr)
    for iE in prange(nE):
        Qi = Q[iE]
        for iX in prange(nX):
            _tr_3333_out(arr[iE, iX, :, :, :, :], Qi, res[iE, iX, :, :, :, :])
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def _tr_3333_multi3(arr: ndarray, Q: ndarray) -> ndarray:
    nE = arr.shape[0]
    res = np.zeros_like(arr)
    for iE in prange(nE):
        _tr_3333_out(arr[iE, :, :, :, :], Q, res[iE, :, :, :, :])
    return res



import numpy as np
from numba import jit
from numba.types import int64, Array
from numba.typed import Dict

from ..linalg.sparse import csr_matrix

__all__ = ["rooted_level_structure", "pseudo_peripheral_nodes"]

int64A = Array(int64, 1, "C")


@jit(nopython=True, nogil=True, fastmath=False, cache=True)
def rooted_level_structure(adj: csr_matrix, root: int = 0) -> Dict:
    """
    Turns a sparse adjacency matrix into a rooted level structure.

    Parameters
    ----------
    adj : csr_matrix
        Adjacency matrix in CSR format.
    root : int, Optional
        Index of the root node. Default is 0.

    Returns
    -------
    dict
        A `numba` dictionary <int[:] : int[:, :]>, where the keys
        refer to different levels, and the values are the indices
        of nodes on that level.
    """
    nN = len(adj.indptr) - 1
    rls = Dict.empty(
        key_type=int64,
        value_type=int64A,
    )
    level = 0
    rls[level] = np.array([root], dtype=np.int64)
    nodes = np.zeros(nN, dtype=np.int64)
    nodes[root] = 1
    levelset = np.zeros(nN, dtype=np.int64)
    nE = 1
    while nE < nN:
        levelset[:] = 0
        for node in rls[level]:
            neighbours = adj.irow(node)
            levelset[neighbours] = 1
        for iN in range(nN):
            if nodes[iN] == 1:
                levelset[iN] = 0
        level += 1
        rls[level] = np.where(levelset == 1)[0]
        nE += len(rls[level])
        for iN in range(nN):
            if levelset[iN] == 1:
                nodes[iN] = 1
    return rls


@jit(nopython=True, nogil=True, fastmath=False, cache=True)
def pseudo_peripheral_nodes(adj: csr_matrix) -> np.ndarray:
    """
    Returns the indices of nodes that are possible candidates
    for being peripheral nodes of a graph.

    Parameters
    ----------
    adj : csr_matrix
        Adjacency matrix in CSR format.

    Returns
    -------
    numpy.ndarray
        Integer array of nodal indices.
    """

    def length_width(RLS):
        length = len(RLS)
        width = 0
        for i in range(length):
            width = max(width, len(RLS[i]))
        return length, width

    RLS = rooted_level_structure(adj, root=0)
    length, width = length_width(RLS)
    while True:
        nodes = RLS[len(RLS) - 1]
        found = False
        for _, node in enumerate(nodes):
            iRLS = rooted_level_structure(adj, root=node)
            iL, iW = length_width(iRLS)
            if (iL > length) or (iL == length and iW < width):
                RLS = iRLS
                length = iL
                width = iW
                found = True
        if not found:
            nR = len(RLS[len(RLS) - 1]) + 1
            res = np.zeros(nR, dtype=np.int64)
            res[:-1] = RLS[len(RLS) - 1]
            res[-1] = RLS[0][0]
            return res



import numpy as np

from sigmaepsilon.core.downloads import download_file, delete_downloads


__all__ = [
    "download_mls_testdata",
    "delete_downloads",
]


def download_mls_testdata() -> np.ndarray:
    """
    Downloads the data of a cloud of points.

    Returns
    -------
    numpy.ndarray
        2 dimensional array of shape (102, 6), which means 102
        points with 6 data per each.

    Example
    --------
    >>> from sigmaepsilon.math.downloads import download_mls_testdata
    >>> download_mls_testdata().shape  # doctest: +SKIP
    (102, 6)

    """
    return np.loadtxt(download_file("mls_testdata.out"))  # pragma: no cover



from os.path import dirname, abspath
from importlib.metadata import metadata

from sigmaepsilon.core.config import namespace_package_name

from .hist import *
from .utils import *
from .decorate import *

__pkg_name__ = namespace_package_name(dirname(abspath(__file__)), 10)
__pkg_metadata__ = metadata(__pkg_name__)
__version__ = __pkg_metadata__["version"]
__description__ = __pkg_metadata__["summary"]
del __pkg_metadata__



import numpy as np
from numpy import ndarray
import scipy as sp
from scipy.spatial import KDTree as KD_scipy
from packaging import version
import warnings

try:  # pragma: no cover
    from sklearn.neighbors import KDTree as KD_sklearn

    __has_sklearn__ = True
except Exception:
    __has_sklearn__ = False
    KD_sklearn = None

__scipy_version__ = sp.__version__


__all__ = ["k_nearest_neighbours"]


def k_nearest_neighbours(
    X: ndarray,
    Y: ndarray | None = None,
    *,
    backend: str = "scipy",
    k: int = 1,
    workers: int = -1,
    tree_kwargs: dict | None = None,
    query_kwargs: dict | None = None,
    leaf_size: int = 30,
    return_distance: bool = False,
    max_distance: float | None = None,
):
    """
    Returns the k nearest neighbours (KNN) of a KDTree for a pointcloud using `scipy`
    or `sklearn`. The function acts as a uniform interface for similar functionality
    of `scipy` and `sklearn`. The most important parameters are highlighted, for the
    complete list of arguments, see the corresponding docs:

    https://docs.scipy.org/doc/scipy/reference/generated/scipy.spatial.KDTree.html#scipy.spatial.KDTree

    https://scikit-learn.org/stable/modules/generated/sklearn.neighbors.KDTree.html

    To learn more about nearest neighbour searches in general:

    https://scikit-learn.org/stable/modules/neighbors.html

    Parameters
    ----------
    X: numpy.ndarray
        An array of points to build the tree.
    Y: numpy.ndarray, Optional
        An array of sampling points to query the tree. If None it is the
        same as the points used to build the tree. Default is None.
    k: int or Sequence[int], Optional
        Either the number of nearest neighbors to return,
        or a list of the k-th nearest neighbors to return, starting from 1.
    leaf_size: positive int, Optional
        The number of points at which the algorithm switches over to brute-force.
        Default is 10.
    workers: int, Optional
        Only if backend is 'scipy'.
        Number of workers to use for parallel processing. If -1 is given all
        CPU threads are used. Default: -1.
        New in 'scipy' version 1.6.0.
    max_distance: float, Optional
        Return only neighbors within this distance. It can be a single value, or
        an array of values of shape matching the input, while a None value
        translates to an infinite upper bound.
        Default is None.
    tree_kwargs: dict, Optional
        Extra keyword arguments passed to the KDTree creator of the selected
        backend. Default is None.

    Returns
    -------
    d: float or array of floats
        The distances to the nearest neighbors. Only returned if
        `return_distance==True`.
    i: integer or array of integers
        The index of each neighbor.

    Raises
    ------
    ImportError
        In the abscence of a usable backend.

    """
    tree_kwargs = {} if tree_kwargs is None else tree_kwargs
    query_kwargs = {} if query_kwargs is None else query_kwargs
    if backend == "scipy":
        tree = KD_scipy(X, leafsize=leaf_size, **tree_kwargs)
        max_distance = np.inf if max_distance is None else max_distance
        query_kwargs["distance_upper_bound"] = max_distance
        if version.parse(__scipy_version__) < version.parse("1.6.0"):
            warnings.warn(
                "Multithreaded execution of a KNN search is "
                + "running on a single thread in scipy<1.6.0. Install a newer"
                + "version or use `backend=sklearn` if scikit is installed."
            )
            d, i = tree.query(Y, k=k, **query_kwargs)
        else:
            d, i = tree.query(Y, k=k, workers=workers)
    elif backend == "sklearn":  # pragma: no cover
        if not __has_sklearn__:
            raise ImportError("'sklearn' must be installed for this!")

        tree = KD_sklearn(X, leaf_size=leaf_size, **tree_kwargs)
        if max_distance is None:
            d, i = tree.query(Y, k=k, **query_kwargs)
        else:
            r = max_distance
            d, i = tree.query_radius(Y, r, k=k, **query_kwargs)
    else:  # pragma: no cover
        raise ImportError("Either `sklearn` or `scipy` must be present for this!")
    return (d, i) if return_distance else i



from typing import TypeVar, Callable, Iterable
from collections import OrderedDict
import warnings

from sympy import Expr, degree, latex
from sympy.core.numbers import One
import numpy as np

from sigmaepsilon.core.kwargtools import getasany

from .metafunction import MetaFunction
from .symutils import substitute


__all__ = ["Function", "FunctionLike"]


FunctionLike = TypeVar("FunctionLike", str, Callable, Expr)


class Function(MetaFunction):
    """
    Base class for all kinds of functions. It can be used to represent
    symbolic and numerical functions. The class is designed to be as
    flexible as possible, so it can be used in a wide range of applications.

    As you can see from the examples, the class can be used in a variety of ways.
    You can initialize a function object with a string expression, a `SymPy` expression,
    or with numerical functions. If you initialize the object with a string or a `SymPy`
    expression, first and second derivatives can be calculated automatically.

    Parameters
    ----------
    f0: FunctionLike or None, Optional
        Positional parameter. A callable object that returns function evaluations.
    f1: Callable or None, Optional
        Positional parameter.
        A callable object that returns evaluations of the gradient of the function.
        If the first positional argument is a string or a `SymPy` expression, this
        parameter can be ignored and derivative functions are calculated automatically.
    f2: Callable or None, Optional
        Positional parameter.
        A callable object that returns evaluations of the Hessian of the function.
        If the first positional argument is a string or a `SymPy` expression, this
        parameter can be ignored and derivative functions are calculated automatically.
    variables: List, Optional
        Symbolic variables. Only required if the function is defined by
        a string or `SymPy` expression.
    value: Callable, Optional
        Keyword only parameter. Same as `f0`.
    gradient: Callable, Optional
        Keyword only parameter.Same as `f1`.
    Hessian: Callable, Optional
        Keyword only parameter.Same as `f2`.
    dimension or dim or d : int, Optional
        Keyword only parameter. The number of dimensions of the domain of the function. This property
        is only required if the function is defined by numerical functions.

    See Also
    --------
    :class:`~sigmaepsilon.math.function.relation.Relation`
    :class:`~sigmaepsilon.math.function.relation.Equality`
    :class:`~sigmaepsilon.math.function.relation.InEquality`

    Examples
    --------
    >>> from sigmaepsilon.math.function import Function
    >>> import sympy as sy
    >>> import numpy as np

    Define a symbolic function with positive variables. Note here that if it was not relevant
    from the aspect of the application to indicate that the variables are positive, it
    wouldn't be necessary to explicity specify them using the parameter `variables`.

    >>> variables = ['x1', 'x2', 'x3', 'x4']
    >>> x1, x2, x3, x4 = syms = sy.symbols(variables, positive=True)
    >>> f = Function(3*x1 + 9*x3 + x2 + x4, variables=syms)
    >>> f([0, 6, 0, 4])
    10

    An equivalent definition can be given without using SymPy explicitly:

    >>> f = Function("3*x1 + 9*x3 + x2 + x4", variables=['x1', 'x2', 'x3', 'x4'])
    >>> f([0, 6, 0, 4])
    10

    In both cases, providing the argument 'variables' is optional, but it determines the
    order of the arguments. If the function is defined without the variables being provided,
    it is derived from the input, but the order of the arguments may differ from what you would
    expect. Also defining the variables with SymPy might be important if you want to indicate
    something about their nature, eg. that they are positive, as in the example above. For instance
    if you execute the following code, the result my change from execution to execution.

    >>> f = Function("3*x1 + 9*x3 + x2 + x4")
    >>> f.variables  # doctest: +SKIP
    (x3, x1, x4, x2)

    Define a numerical function. In this case the dimension of the input must be specified
    explicitly.

    >>> def f0(x, y): return x**2 + y
    >>> def f1(x, y): return np.array([2*x, 1])
    >>> def f2(x, y): return np.array([[0, 0], [0, 0]])
    >>> f = Function(f0, f1, f2, d=2)
    >>> f.is_linear
    False

    To call the function, call it like you would call the function `f0`:

    >>> f(1, 1)
    2
    >>> f.g(1, 1)
    array([2, 1])

    You can mix different kinds of input signatures:

    >>> def f0(x): return x[0]**2 + x[1]
    >>> def f1(x, y): return np.array([2*x, 1])
    >>> def f2(x, y): return np.array([[0, 0], [0, 0]])
    >>> f = Function(f0, f1, f2, d=2)

    The point is that you always call the resulting `Function` object
    according to your definition. Now your `f0` expects an iterable,
    therefore you can call it like

    >>> f([1, 1])
    2

    but the gradient function expects the same values as two scalars, so
    you call it like this

    >>> f.g(1, 1)
    array([2, 1])

    Explicity defining the variables for a symbolic function is important
    if not all variables appear in the string expression you feed the object with:

    >>> g = Function('3*x + 4*y - 2', variables=['x', 'y', 'z'])
    >>> g.is_linear
    True

    If you do not specify 'z' as a variable here, the resulting object expects
    two values

    >>> h = Function('3*x + 4*y - 2')
    >>> h([1, 1])
    5

    The variables can be `SymPy` variables as well:

    >>> m = Function('3*x + 4*y - 2', variables=sy.symbols('x y z'))
    >>> m.is_linear
    True
    >>> m([1, 2, -30])
    9

    """

    __slots__ = ("f0", "f1", "f2", "dimension", "domain", "vmap", "from_str")

    # NOTE domain is missing from the possible parameters
    # NOTE investigate if dimensions could be derived

    def __init__(
        self,
        f0: FunctionLike | None = None,
        f1: Callable | None = None,
        f2: Callable | None = None,
        *args,
        variables: Iterable | None = None,
        **kwargs
    ):
        super().__init__()
        self.update(f0, f1, f2, *args, variables=variables, **kwargs)

    def update(
        self,
        f0: FunctionLike | None = None,
        f1: Callable | None = None,
        f2: Callable | None = None,
        *_,
        variables: Iterable | None = None,
        **kwargs
    ):
        self.from_str = None

        if f0 is not None:
            if isinstance(f0, str):
                kwargs.update(self._str_to_func(f0, variables=variables, **kwargs))
                self.from_str = True
            elif isinstance(f0, Expr):
                kwargs.update(self._sympy_to_func(f0, variables=variables, **kwargs))

        self.expr = kwargs.get("expr", None)
        self.variables = kwargs.get("variables", variables)
        self.f0 = kwargs.get("value", f0)
        self.f1 = kwargs.get("gradient", f1)
        self.f2 = kwargs.get("Hessian", f2)
        self.dimension = getasany(["d", "dimension", "dim"], None, **kwargs)
        self.domain = kwargs.get("domain", None)
        self.vmap = kwargs.get("vmap", None)

    @property
    def symbolic(self) -> bool:
        """
        Returns `True` if the function is a fit subject of symbolic manipulation.
        This is probably only true if the object was created from a string or
        `sympy` expression.
        """
        warnings.warn(
            "The property `symbolic` is deprecated and will be removed in a future version. "
            "Use `is_symbolic` instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.is_symbolic

    @property
    def linear(self) -> bool:
        """
        Returns True if the function is at most linear in all of its variables.
        """
        warnings.warn(
            "The property `linear` is deprecated and will be removed in a future version. "
            "Use `is_linear` instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.is_linear

    @property
    def is_linear(self) -> bool:
        """
        Returns `True` if the function is at most linear in all of its variables.
        """
        if self.is_symbolic:
            return all(
                np.array([degree(self.expr, v) for v in self.variables], dtype=int) <= 1
            )
        else:
            return self.f2 is None

    @property
    def is_symbolic(self) -> bool:
        """
        Returns `True` if the function is a fit subject of symbolic manipulation.
        This is probably only true if the object was created from a string or
        `SymPy` expression.
        """
        return self.expr is not None

    def simplify(self) -> None:
        """
        Simplifies the symbolic expression of the instance.
        """
        if self.is_symbolic:
            self.expr = self.expr.simplify()
        else:
            raise TypeError("This is exclusive to symbolic functions.")

    def linear_coefficients(self, normalize: bool = False) -> dict | None:
        """
        Returns the linear coeffiecients, if the function is symbolic and linear.

        Parameters
        ----------
        normalize : bool, Optional
            If True, the coefficients are normalized. Default is False.

        Examples
        --------
        >>> from sigmaepsilon.math.function import Function
        >>> from sigmaepsilon.math.approx.lagrange import gen_Lagrange_1d
        >>> f = gen_Lagrange_1d(N=2)
        >>> f1 = Function(f[1][0], f[1][1], f[1][2])
        >>> linear_coefficients = f1.linear_coefficients()

        """
        d = self.coefficients(normalize)
        if d:
            return {
                key: value for key, value in d.items() if len(key.free_symbols) <= 1
            }
        return None

    def coefficients(self, normalize: bool = False) -> dict | None:
        """
        Returns the coefficients if the function is symbolic.

        Parameters
        ----------
        normalize : bool, Optional
            If True, the coefficients are normalized. Default is False.

        Examples
        --------
        >>> from sigmaepsilon.math.function import Function
        >>> from sigmaepsilon.math.approx.lagrange import gen_Lagrange_1d
        >>> f = gen_Lagrange_1d(N=2)
        >>> f1 = Function(f[1][0], f[1][1], f[1][2])
        >>> coefficients = f1.coefficients()

        """
        if not self.is_symbolic:
            raise TypeError("This is exclusive to symbolic functions.")

        try:
            d = OrderedDict({x: 0 for x in self.variables})
            expr = self.expr.simplify()
            d.update(expr.as_coefficients_dict())
            if not normalize:
                return d
            else:
                res = OrderedDict()
                for key, value in d.items():
                    if len(key.free_symbols) == 0:
                        res[One()] = value * key
                    else:
                        res[key] = value
                return res
        except Exception:  # pragma: no cover
            return None

    def to_latex(self) -> str:
        """
        Returns the LaTeX code of the symbolic expression of the instance.
        Only for symbolic functions.

        Examples
        --------
        >>> from sigmaepsilon.math.function import Function
        >>> from sigmaepsilon.math.approx.lagrange import gen_Lagrange_1d
        >>> f = gen_Lagrange_1d(N=2)
        >>> f1 = Function(f[1][0], f[1][1], f[1][2])
        >>> latex_string = f1.to_latex()

        """
        if self.is_symbolic:
            return latex(self.expr)
        else:
            raise TypeError("This is exclusive to symbolic functions.")

    def subs(
        self, values: Iterable, variables: Iterable | None = None, inplace: bool = False
    ) -> "Function":
        """
        Substitites values for variables. Only for symbolic functions.

        Examples
        --------
        >>> from sigmaepsilon.math.function import Function
        >>> g = Function("3*x + 4*y - 2", variables=["x", "y", "z"])
        >>> g = g.subs([0, 0, 0], ["x", "y", "z"], inplace=True)

        """
        if not self.is_symbolic:
            raise TypeError("This is exclusive to symbolic functions.")

        expr = substitute(self.expr, values, variables, as_string=self.from_str)
        kwargs = self._sympy_to_func(expr=expr, variables=variables)

        if not inplace:
            return Function(None, None, None, **kwargs)
        else:
            self.update(None, None, None, **kwargs)
            return self



from typing import Union
import numpy as np
from numpy import ndarray
from .utils import atleast1d


def is_none_or_false(a):
    if isinstance(a, bool):
        return not a
    elif a is None:
        return True
    return False


def isfloatarray(a: ndarray) -> bool:
    """
    Returns `True` if `a` is a float array.
    """
    return np.issubdtype(a.dtype, float)


def isintegerarray(a: ndarray) -> bool:
    """
    Returns `True` if `a` is an integer array.
    """
    return np.issubdtype(a.dtype, int)


def isintarray(a: ndarray) -> bool:
    """
    Returns `True` if `a` is a integer array.
    """
    return isintegerarray(a)


def isboolarray(a: ndarray) -> bool:
    """
    Returns `True` if `a` is a boolean array.
    """
    return np.issubdtype(a.dtype, bool)


def is1dfloatarray(a: ndarray) -> bool:
    """
    Returns `True` if `a` is a 1d float array.
    """
    return isfloatarray(a) and len(a.shape) == 1


def is1dintarray(a: ndarray) -> bool:
    """
    Returns `True` if `a` is a 1d integer array.
    """
    return isintarray(a) and len(a.shape) == 1


def issymmetric(a: ndarray, tol: float = 1e-8) -> bool:
    """
    Returns `True` if `a` is symmetric with a given tolerance
    prescribed by `tol`.
    """
    return np.linalg.norm(a - a.T) < tol


def isposdef(A: ndarray, tol=0) -> bool:
    """
    Returns `True` if `A` is positive definite.

    Examples
    --------
    >>> from sigmaepsilon.math.linalg import random_posdef_matrix
    >>> from sigmaepsilon.math.logical import isposdef
    ...
    >>> A = random_posdef_matrix(3, 0.1)
    >>> isposdef(A)
    True

    >>> A[0, 0] = 0
    >>> isposdef(A)
    False

    """
    return np.all(np.linalg.eigvals(A) > tol)


def ispossemidef(A: ndarray) -> bool:
    """
    Returns `True` if `A` is positive semidefinite.

    Example
    -------
    >>> from sigmaepsilon.math.linalg import random_pos_semidef_matrix
    >>> from sigmaepsilon.math.logical import ispossemidef
    >>> A = random_pos_semidef_matrix(3)
    >>> ispossemidef(A)
    True

    """
    return np.all(np.linalg.eigvals(A) >= 0)


def isclose(
    x1: ndarray, x2: ndarray, *, atol: float = 1e-8, rtol: float = 1e-5
) -> Union[bool, ndarray]:
    """
    Returns a boolean array where two arrays are element-wise equal
    in absolute and optionally in relative sense. In the latter case
    the relative difference is measured agains `x2`.

    Absolute difference is measured as abs(`x1` - `x2`) <= `atol`, relative
    difference is measured as abs(`x1` - `x2`) <= abs(`rtol` * `x2`).

    .. versionadded:: 0.0.6

    .. warning:: Default values might fail for very small numbers.

    Parameters
    ----------
    x1 : ndarray or float
        Values of the first array.
    x2 : ndarray or float
        Values of the second array.
    atol : float, Optional
        Absolute tolerance. It must be a positive number, or None to
        turn this check off. Default is 1e-8.
    rtol : float, Optional
        Relative tolerance. It must be a positive number or None to
        turn this check off. In the latter case, the input arrays are only
        compared in an absolute sense. Default is 1e-5.

    Returns
    -------
    numpy.ndarray or float
        A boolean array or a float, depending on the input.

    See Also
    --------
    numpy.isclose
    numpy.allclose
    math.isclose
    """
    if not isinstance(x1, ndarray):
        x1 = atleast1d(x1)
    if not isinstance(x2, ndarray):
        x2 = atleast1d(x2)
    c = np.ones(x1.shape, dtype=bool)
    assert (atol is not None) or (
        rtol is not None
    ), "At least one from `atol` or `rtol` must be other than `None`."
    if atol is not None:
        assert atol > 0, "The absolute tolerance must be a positive number."
        c_abs = np.abs(x1 - x2) <= atol
        c = c & c_abs
    if rtol is not None:
        assert rtol > 0, "The relative tolerance must be a positive number."
        c_rel = np.abs(x1 - x2) <= np.abs(rtol * x2)
        c = c & c_rel
    if len(c.shape) == 1 and c.shape[0] == 1:
        return c[0]
    return c


def allclose(*args, **kwargs) -> bool:
    """
    Same as `isclose`, but it returns a single boolean that is `True` or `False`
    if all the values returned by `isclose` are `True` or `False`.

    See Also
    --------
    isclose
    """
    return np.all(isclose(*args, **kwargs))



from typing import TypeAlias, Sequence

__all__ = ["Scalar", "BoundLike", "BoundsLike"]

Scalar: TypeAlias = int | float
# type NumberLike = int | float  # after Python 3.12 and higher

BoundLike: TypeAlias = tuple[Scalar | None, Scalar | None]

BoundsLike: TypeAlias = BoundLike | Sequence[BoundLike]



from .utils import *
from .frame import *
from .vector import *
from .tensor import *
from .meta import *
from .logical import *
from .sparse import JaggedArray, csr_matrix

__all__ = ["JaggedArray", "csr_matrix"]

from .utils import __all__ as _utils_all
from .frame import __all__ as _frame_all
from .vector import __all__ as _vector_all
from .tensor import __all__ as _tensor_all
from .meta import __all__ as _meta_all
from .logical import __all__ as _logical_all

__all__.extend(_utils_all)
__all__.extend(_frame_all)
__all__.extend(_vector_all)
__all__.extend(_tensor_all)
__all__.extend(_meta_all)
__all__.extend(_logical_all)



import numpy as np
from numpy import ndarray
from awkward import Array as akarray
from numba import njit
from numba.core import types
from numba.typed import Dict
from typing import Union, Any
import awkward as ak

from .utils import flatten2dC
from .utils import count_cols

__cache = True


__all__ = ["unique2d"]


ArrayLike = Union[ndarray, akarray]
i64 = types.int64
i64A = types.int64[:]
i64A2 = types.int64[:, :]


def unique2d(
    arr: ArrayLike | Any,
    return_index: bool = False,
    return_inverse: bool = False,
    return_counts: bool = False,
):
    """
    Find the unique elements of a 2d array.

    Returns the sorted unique elements of an array. There are three optional
    outputs in addition to the unique elements:

    * the indices of the input array that give the unique values
    * the indices of the unique array that reconstruct the input array
    * the number of times each unique value comes up in the input array

    Parameters
    ----------
    arr : numpy.ndarray
        Input array.
    return_index : bool, optional
        If True, also return the indices of `ara` that result in the unique array.
    return_inverse : bool, optional
        If True, also return the indices of the unique array that can be used to
        reconstruct `arr`.
    return_counts : bool, optional
        If True, also return the number of times each unique item appears
        in `arr`.

    Example
    -------
    >>> from sigmaepsilon.math.arraysetops import unique2d
    >>> import numpy as np
    >>> arr = np.array([[1, 2, 3], [1, 2, 4]], dtype=int)
    >>> unique2d(arr)
    array([1, 2, 3, 4])

    >>> unique2d(arr, return_index=True)
    [array([1, 2, 3, 4]), DictType[int64,array(int64, 2d, A)]<iv=None>({1: [[0 0]
     [1 0]], 2: [[0 1]
     [1 1]], 3: [[0 2]], 4: [[1 2]]})]

    >>> unique2d(arr, return_inverse=True)  # doctest: +SKIP
    [array([1, 2, 3, 4]), array([[0, 1, 2],
           [0, 1, 3]])]

    >>> unique2d(arr, return_counts=True)  # doctest: +SKIP
    [array([1, 2, 3, 4]), array([2, 2, 1, 1])]

    >>> from sigmaepsilon.math.linalg.sparse import JaggedArray
    >>> arr = JaggedArray(np.array([1, 2, 1, 2, 3]), cuts=[2, 3])
    >>> unique2d(arr)
    array([1, 2, 3])

    """
    if isinstance(arr, ndarray):
        unique, counts, inverse, indices = _unique2d_njit(arr)
        res = [
            unique,
        ]
        if return_index:
            res.append(indices)
        if return_inverse:
            res.append(inverse)
        if return_counts:
            res.append(counts)
        return res if len(res) > 1 else res[0]
    elif isinstance(arr, akarray):
        assert arr.ndim == 2, "Only 2 dimensional awkward arrays are supported!"
        flatarr = ak.flatten(arr).to_numpy()
        unique, counts, inverse, indices = _unique1d_njit(flatarr)
        res = [
            unique,
        ]
        if return_index:
            res.append(indices)
        if return_inverse:
            res.append(ak.unflatten(inverse, count_cols(arr)))
        if return_counts:
            res.append(counts)
        return res if len(res) > 1 else res[0]
    else:
        if hasattr(arr, "is_jagged"):
            if not arr.is_jagged():
                return unique2d(
                    arr.to_numpy(),
                    return_index=return_index,
                    return_inverse=return_inverse,
                    return_counts=return_counts,
                )
            else:
                return unique2d(
                    arr.to_ak(),
                    return_index=return_index,
                    return_inverse=return_inverse,
                    return_counts=return_counts,
                )
    raise TypeError(f"Invalid input type {type(arr)}")


@njit(nogil=True, parallel=False, fastmath=False, cache=__cache)
def _unique1d_njit(flatdata: ndarray):
    unique = np.unique(flatdata)
    nF = len(flatdata)
    nU = len(unique)
    counts = np.zeros(nU, dtype=np.int64)
    inverse = np.zeros(nF, dtype=np.int64)
    imap = Dict.empty(key_type=i64, value_type=i64)
    for i in range(nU):
        imap[unique[i]] = i
    for i in range(nF):
        counts[imap[flatdata[i]]] += 1
        inverse[i] = imap[flatdata[i]]
    indices = Dict.empty(key_type=i64, value_type=i64A)
    for i in range(nU):
        indices[unique[i]] = np.zeros((counts[i]), dtype=np.int64)
    counts[:] = 0
    for iF in range(nF):
        i = inverse[iF]
        indices[unique[i]][counts[i]] = iF
        counts[i] += 1
    return unique, counts, inverse, indices


@njit(nogil=True, parallel=False, fastmath=False, cache=__cache)
def _unique2d_njit(data: ndarray):
    flatdata = flatten2dC(data)
    unique = np.unique(flatdata)
    nF = len(flatdata)
    nU = len(unique)
    counts = np.zeros(nU, dtype=np.int64)
    inverse = np.zeros(nF, dtype=np.int64)
    imap = Dict.empty(key_type=i64, value_type=i64)
    for i in range(nU):
        imap[unique[i]] = i
    for i in range(nF):
        counts[imap[flatdata[i]]] += 1
        inverse[i] = imap[flatdata[i]]
    inverse = inverse.reshape(data.shape)
    indices = Dict.empty(key_type=i64, value_type=i64A2)
    for i in range(nU):
        indices[unique[i]] = np.zeros((counts[i], 2), dtype=np.int64)
    counts[:] = 0
    nR, nC = data.shape
    for r in range(nR):
        for c in range(nC):
            i = inverse[r, c]
            indices[unique[i]][counts[i], 0] = r
            indices[unique[i]][counts[i], 1] = c
            counts[i] += 1
    return unique, counts, inverse, indices



from typing import Iterable

from .function import Function


class TestFunction(Function):
    __slots__ = ("optimums", "optText")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.optimums = []

        self.optText = "opt"
        for key, value in kwargs.items():
            if key == "optText":
                assert isinstance(value, str)
                self.optText = value
                break
            elif key == "optimums":
                assert isinstance(value, Iterable)
                for v in value:
                    self.optimums.append(v)
        return


class TestFunction2D(TestFunction):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class TestMinFunction(TestFunction):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, optText="min")


class TestMaxFunction(TestFunction):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs, optText="max")


class TestMinFunction2D(TestMinFunction, TestFunction2D): ...


class TestMaxFunction2D(TestMaxFunction, TestFunction2D): ...



from typing import Tuple
import numpy as np
from numpy import ndarray

from .ga import GeneticAlgorithm

__all__ = ["BinaryGeneticAlgorithm"]


class BinaryGeneticAlgorithm(GeneticAlgorithm):
    """
    An implementation of a Binary Genetic Algorithm (BGA) for finding
    minimums of real valued unconstrained problems of continuous variables
    in n-dimensional vector spaces.

    The class is able to solve unconstrained optimization problems of the form:

    .. math::

        \\begin{eqnarray}
            & minimize&  \\quad  f(\\mathbf{x}) \\quad in \\quad \\mathbf{x} \\in \\mathbf{R}^n.
        \\end{eqnarray}

    Parameters
    ----------
    fnc: Callable
        The function to evaluate. It is assumed, that the function expects and N
        number of scalar arguments as a 1d iterable.
    ranges: Iterable
        Ranges for each scalar argument to the objective function.
    length: int, Optional
        Chromosome length. The higher the value, the more precision. Default is 5.
    p_c: float, Optional
        Probability of crossover. Default is 1.
    p_m: float, Optional
        Probability of mutation. Default is 0.2.
    nPop: int, Optional
        The size of the population. Default is 100.
    maxiter: int, Optional
        The maximum number of iterations. Default is 200.
    miniter: int, Optional
        The minimum number of iterations. Default is 100.
    elitism: float or int, Optional
        Default is 1
    ftol: float, Optional
        Torelance for floating point operations. Default is 1e-12.
    maxage: int, Optional
        The age is the number of generations a candidate spends at the top
        (being the best candidate). Setting an upper limit to this value is a kind
        of stopping criterion. Default is 5.

    See Also
    --------
    :class:`~sigmaepsilon.math.optimize.ga.GeneticAlgorithm`
    :class:`~sigmaepsilon.math.optimize.ga.Genom`

    Examples
    --------
    Find the minimizer of the Rosenbrock function.
    The exact value of the solution is x = [1.0, 1.0].

    >>> from sigmaepsilon.math.optimize import BinaryGeneticAlgorithm as BGA
    >>>
    >>> def Rosenbrock(x):
    ...     a, b = 1, 100
    ...     return (a-x[0])**2 + b*(x[1]-x[0]**2)**2
    >>>
    >>>
    >>> ranges = [[-10, 10], [-10, 10]]
    >>> bga = BGA(Rosenbrock, ranges, length=12, nPop=200)
    >>> _ = bga.solve()
    >>> x = bga.best_phenotype()
    >>> fx = Rosenbrock(x)

    The following code prints the history using the :func:`evolve` generator method of
    the instance

    >>> bga = BGA(Rosenbrock, ranges, length=12, nPop=200)
    >>> _ = [bga.evolve(1) for _ in range(100)]
    >>> x = bga.best_phenotype()
    >>> fx = Rosenbrock(x)

    """

    def populate(self, genotypes: ndarray | None = None) -> ndarray:
        """
        Populates the model and returns the array of genotypes.
        """
        nPop = self.nPop

        if genotypes is None:
            poolshape = (int(nPop / 2), self.dim * self.length)
            genotypes = np.random.randint(2, size=poolshape)
        else:
            poolshape = genotypes.shape

        nParent = poolshape[0]

        if nParent < nPop:
            offspring = []
            g = self.random_parents_generator(genotypes)
            try:
                while (len(offspring) + nParent) < nPop:
                    parent1, parent2 = next(g)
                    offspring.extend(self.crossover(parent1, parent2))
                genotypes = np.vstack([genotypes, offspring])
            except Exception:  # pragma: no cover
                raise RuntimeError

        return genotypes

    def decode(self, genotypes: ndarray) -> ndarray:
        """
        Decodes the genotypes to phenotypes and returns them as an array.
        """
        span = 2**self.length - 2**0
        genotypes = genotypes.reshape((self.nPop, self.dim, self.length))
        precisions = [
            (self.ranges[d, -1] - self.ranges[d, 0]) / span for d in range(self.dim)
        ]
        phenotypes = np.sum(
            [genotypes[:, :, i] * 2**i for i in range(self.length)], axis=0
        ).astype(float)
        for d in range(self.dim):
            phenotypes[:, d] *= precisions[d]
            phenotypes[:, d] += self.ranges[d, 0]
        return phenotypes

    def crossover(
        self, parent1: ndarray, parent2: ndarray, nCut: int | None = None
    ) -> Tuple[ndarray]:
        """
        Performs crossover on the parents `parent1` and `parent2`,
        using an `nCut` number of cuts and returns two childs.
        """
        if np.random.rand() > self.p_c:
            return parent1, parent2

        if nCut is None:
            nCut = np.random.randint(1, self.dim * self.length - 1)

        cuts = [0, self.dim * self.length]
        p = np.random.choice(range(1, self.length * self.dim - 1), nCut, replace=False)
        cuts.extend(p)
        cuts = np.sort(cuts)

        child1 = np.zeros(self.dim * self.length, dtype=int)
        child2 = np.zeros(self.dim * self.length, dtype=int)

        randBool = np.random.rand() > 0.5
        for i in range(nCut + 1):
            if (i % 2 == 0) == randBool:
                child1[cuts[i] : cuts[i + 1]] = parent1[cuts[i] : cuts[i + 1]]
                child2[cuts[i] : cuts[i + 1]] = parent2[cuts[i] : cuts[i + 1]]
            else:
                child1[cuts[i] : cuts[i + 1]] = parent2[cuts[i] : cuts[i + 1]]
                child2[cuts[i] : cuts[i + 1]] = parent1[cuts[i] : cuts[i + 1]]

        return self.mutate(child1), self.mutate(child2)

    def mutate(self, child: ndarray) -> ndarray:
        """
        Returns a mutated genotype. Children come in, mutants go out.
        """
        p = np.random.rand(self.dim * self.length)
        return np.where(p > self.p_m, child, 1 - child)

    def select(self, genotypes: ndarray, phenotypes: ndarray | None = None) -> ndarray:
        """
        Organizes a tournament and returns the genotypes of the winners.
        """
        fittness = self.evaluate(phenotypes)
        winners, others = self.divide(fittness)
        while len(winners) < int(self.nPop / 2):
            candidates = np.random.choice(others, 3, replace=False)
            winner = np.argsort([fittness[ID] for ID in candidates])[0]
            winners.append(candidates[winner])
        return np.array([genotypes[w] for w in winners], dtype=float)



import numpy as np
import numbers
from numpy import ndarray
from numba import njit, prange
from typing import Union, Tuple, Iterable

try:
    from collections.abc import Iterable  # pragma: no cover
except ImportError:  # pragma: no cover
    from collections import Iterable  # pragma: no cover

__cache = True

ArrayOrFloat = Union[float, ndarray, list]


def itype_of_ftype(dtype):
    """
    Returns a matching NumPy integer type to a float type.
    """
    name = np.dtype(dtype).name
    if "32" in name:
        return np.int32
    elif "64" in name:
        return np.int64
    else:
        raise TypeError("Unrecognized float type.")  # pragma: no cover


@njit(nogil=True, cache=__cache)
def minmax(a: ndarray) -> Tuple[float]:
    """
    Returns the minimum and maximum values of an array.
    """
    return a.min(), a.max()


def ascont(array: ndarray) -> ndarray:
    """
    Returns the input as contiguous array.
    It is basically a shortcut to `numpy.ascontiguousarray`.
    """
    return np.ascontiguousarray(array)


@njit(nogil=True, cache=__cache)
def clip1d(a: ndarray, a_min: float, a_max: float) -> ndarray:
    """
    Clips the values outside the interval [a_min, a_max] to
    either a_min or a_max.

    Parameters
    ----------
    a : numpy.ndarray
        An 1d array.
    a_min : float
        The lower limit.
    a_max : float
        The upper limit.

    Examples
    --------
    >>> from sigmaepsilon.math import clip1d
    >>> import numpy as np
    ...
    >>> clip1d(np.array([0.9, 2.5]), 1.0, 1.5)
    array([1. , 1.5])

    """
    a[a < a_min] = a_min
    a[a > a_max] = a_max
    return a


def atleastnd(
    a: Union[numbers.Number, Iterable],
    n: int = 2,
    front: bool = True,
    back: bool = False,
) -> ndarray:
    """
    Returns an array that is at least 'n' dimensional.
    The required shape is obtained by inserting new axes either
    before or after existing ones. This behaviour can be controlled
    using the parameters 'front' and 'back'. If front is True and back
    is False, new axes are crated before the first existing data index,
    and the opposite happens in every other case.

    Examples
    --------
    >>> from sigmaepsilon.math import atleastnd
    >>> import numpy as np
    ...
    >>> atleastnd(np.array([1, 1]), 3, front=True).shape
    (1, 1, 2)

    >>> atleastnd(np.array([1, 1]), 3, back=True).shape
    (2, 1, 1)

    """
    if not isinstance(a, Iterable):
        a = [
            a,
        ]
    if not isinstance(a, ndarray):
        a = np.array(a)
    shp = a.shape
    nD = len(shp)
    if nD >= n:
        return a
    else:
        if front and not back:
            newshape = (n - nD) * (1,) + shp
        else:
            newshape = shp + (n - nD) * (1,)
        return np.reshape(a, newshape)


def atleast1d(a: Union[numbers.Number, Iterable]) -> ndarray:
    """
    Returns an array that is at least 1 dimensional.

    Examples
    --------
    >>> from sigmaepsilon.math import atleast1d
    >>> atleast1d(1)
    array([1])

    """
    return np.atleast_1d(a)


def atleast2d(a: Union[numbers.Number, Iterable], **kwargs) -> ndarray:
    """
    Returns an array that is at least 2 dimensional.

    Examples
    --------
    >>> from sigmaepsilon.math import atleast2d
    >>> atleast2d(1)
    array([[1]])

    """
    return atleastnd(a, 2, **kwargs)


def matrixform(a: Union[numbers.Number, Iterable]) -> ndarray:
    """
    Returns an array that is at least 2 dimensional.

    Examples
    --------
    >>> from sigmaepsilon.math import matrixform
    >>> matrixform(1)
    array([[1]])

    """
    if not isinstance(a, Iterable):
        a = [
            a,
        ]
    if not isinstance(a, ndarray):
        a = np.array(a)
    size = len(a.shape)
    assert size <= 2, "Input array must be at most 2 dimensional."
    if size == 1:
        nV, nC = len(a), 1
        return a.reshape(nV, nC)
    return a


def atleast3d(a: Union[numbers.Number, Iterable], **kwargs) -> ndarray:
    """
    Returns an array that is at least 3 dimensional.

    Examples
    --------
    >>> from sigmaepsilon.math import atleast3d
    >>> atleast3d(1)
    array([[[1]]])

    """
    return atleastnd(a, 3, **kwargs)


def atleast4d(a: ndarray, **kwargs) -> ndarray:
    """
    Returns an array that is at least 4 dimensional.

    Examples
    --------
    >>> from sigmaepsilon.math import atleast4d
    >>> atleast4d(1)
    array([[[[1]]]])

    """
    return atleastnd(a, 4, **kwargs)


@njit(nogil=True, cache=__cache)
def flatten2dC(a: ndarray) -> ndarray:
    I, J = a.shape
    res = np.zeros(I * J, dtype=a.dtype)
    ind = 0
    for i in range(I):
        for j in range(J):
            res[ind] = a[i, j]
            ind += 1
    return res


@njit(nogil=True, cache=__cache)
def flatten2dF(a: ndarray) -> ndarray:
    I, J = a.shape
    res = np.zeros(I * J, dtype=a.dtype)
    ind = 0
    for j in range(J):
        for i in range(I):
            res[ind] = a[i, j]
            ind += 1
    return res


def flatten2d(a: ndarray, order: str = "C") -> ndarray:
    """
    Returns a flattened view of `a`.
    """
    if order == "C":
        return flatten2dC(a)
    elif order == "F":
        return flatten2dF(a)


def bool_to_float(a: Iterable, true: float = 1.0, false: float = 0.0) -> ndarray:
    """
    Transforms a boolean array to a float array using the specified
    values for `True` and `False`.

    Example
    -------
    >>> from sigmaepsilon.math import bool_to_float
    >>> bool_to_float([True, False], 1.0, -2.0)
    array([ 1., -2.])

    """
    if not isinstance(a, ndarray):
        a = np.array(a)
    res = np.full(a.shape, false, dtype=float)
    res[a] = true
    return res


def choice(choices: Iterable, size: Tuple, probs: Iterable = None) -> ndarray:
    """
    Returns a NumPy array, whose elements are selected from
    'choices' under probabilities provided with 'probs' (optionally).

    Parameters
    ----------
    choices : Iterable
        The choices.
    size : tuple
        The size of the output array.
    probs : Iterable, Optional
        The probabilities of each item in 'choices'. Default is an array
        of uniform probabilities.

    Example
    -------
    ```python
    N, p = 2, 0.2
    choice([False, True], (N, N), [p, 1-p])
    array([[ True,  True],
           [ True,  True]])
    ```

    """
    if probs is None:
        probs = np.full((len(choices),), 1 / len(choices))
    return np.random.choice(a=choices, size=size, p=probs)


@njit(nogil=True, parallel=True, cache=__cache)
def repeat(a: ndarray, N: int = 1) -> ndarray:
    """
    Repeats an array N-times.

    Parameters
    ----------
    a: numpy.ndarray
        Input array.
    N: int, Optional
        Number of repetitions. Default is 1.

    Returns
    -------
    numpy.ndarray
        A NumPy array with shape (N, a.shape[0], a.shape[1]).

    Example
    -------
    For example, to generate basis vectors for 10 vectors embedded
    in the same coordinate frame, we need to stack up 10 identical
    identity matrices. This can be done the quickest by:

    >>> from sigmaepsilon.math import repeat
    >>> import numpy as np
    ...
    >>> repeat(np.eye(2), 3)
    array([[[1., 0.],
            [0., 1.]],
    <BLANKLINE>
           [[1., 0.],
            [0., 1.]],
    <BLANKLINE>
           [[1., 0.],
            [0., 1.]]])

    >>> repeat(np.eye(2), 3).shape
    (3, 2, 2)

    """
    res = np.zeros((N, a.shape[0], a.shape[1]), dtype=a.dtype)
    for i in prange(N):
        res[i, :, :] = a
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def repeat1d(a: ndarray, N=1) -> ndarray:
    """
    Repeats a 1d array N times.

    Example
    -------
    >>> from sigmaepsilon.math import repeat1d
    >>> import numpy as np
    ...
    >>> repeat1d(np.array([1, 2]), 3)
    array([1, 2, 1, 2, 1, 2])

    """
    M = a.shape[0]
    res = np.zeros(N * M, dtype=a.dtype)
    for i in prange(N):
        res[i * M : (i + 1) * M] = a
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def tile(a: ndarray, da: ndarray, N: int = 1) -> ndarray:
    """
    Tiles a 1d array N times.

    Example
    -------
    >>> from sigmaepsilon.math import tile
    >>> import numpy as np
    ...
    >>> arr = tile(np.array([[0, 0]]), np.array([[1, -1]]), 3)
    >>> arr
    array([[[ 0,  0]],
    <BLANKLINE>
           [[ 1, -1]],
    <BLANKLINE>
           [[ 2, -2]]])

    """
    res = np.zeros((N, a.shape[0], a.shape[1]), dtype=a.dtype)
    for i in prange(N):
        res[i, :, :] = a + i * da
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def tile1d(a: ndarray, da: ndarray, N=1) -> ndarray:
    M = a.shape[0]
    res = np.zeros(N * M, dtype=a.dtype)
    for i in prange(N):
        res[i * M : (i + 1) * M] = a + i * da
    return res


def indices_of_equal_rows(x: ndarray, y: ndarray, tol=1e-12):
    from .logical import isintegerarray

    nX, dX = x.shape
    nY, dY = y.shape
    assert dX == dY, "Input arrays must have identical second dimensions."
    square = nX == nY
    integer = isintegerarray(x) and isintegerarray(y)
    if integer:
        if square:
            inds = np.flatnonzero((x == y).all(1))
            return inds, np.copy(inds)
        else:
            return indices_of_equal_rows_njit(x, y, 0)
    else:
        if square:
            return indices_of_equal_rows_square_njit(x, y, tol)
        else:
            return indices_of_equal_rows_njit(x, y, tol)


@njit(nogil=True, parallel=True, cache=__cache)
def indices_of_equal_rows_njit(x: ndarray, y: ndarray, tol: float = 1e-12):
    R = np.zeros((x.shape[0], y.shape[0]), dtype=x.dtype)
    for i in prange(R.shape[0]):
        for j in prange(R.shape[1]):
            R[i, j] = np.sum(np.abs(x[i] - y[j]))
    return np.where(R <= tol)


@njit(nogil=True, parallel=True, cache=__cache)
def indices_of_equal_rows_square_njit(x: ndarray, y: ndarray, tol=1e-12):
    nx, ny = x.shape[0], y.shape[0]
    if nx < ny:
        n = nx
    else:
        n = ny
    R = np.zeros((n, n), dtype=x.dtype)
    for i in prange(n):
        for j in prange(n):
            R[i, j] = np.sum(np.abs(x[i] - y[j]))
    return np.where(R <= tol)


@njit(nogil=True, parallel=True, fastmath=True, cache=__cache)
def count_cols(arr: ndarray) -> ndarray:
    """
    Count and return the number of columns for each row in
    the input array.
    """
    n = len(arr)
    res = np.zeros(n, dtype=np.int64)
    for i in prange(n):
        res[i] = len(arr[i])
    return res


@njit(nogil=True, parallel=True, fastmath=True, cache=__cache)
def count_cols_csr(indptr: ndarray, indices: ndarray) -> ndarray:
    """
    Count and return the number of columns for each row in
    a CSR matrix.

    Parameters
    ----------
    indptr : numpy.ndarray
        Index pointers of a CSR matrix.
    indices : numpy.ndarray
        Column indices of the data in a CSR matrix.

    Returns
    -------
    numpy.ndarray
        1d integer array.

    See Also
    --------
    :class:`~sigmaepsilon.math.linalg.sparse.csr.csr_matrix`
    """
    n = len(indptr)
    res = np.zeros(n, dtype=indptr.dtype)
    for i in prange(n):
        res[i] = indices[indptr[i + 1]] - indices[indptr[i]]
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def repeat_diagonal_2d(a: ndarray, N: int = 2) -> ndarray:
    """
    Assembles a 2d dense block-diagonal matrix by repeating.

    Parameters
    ----------
    a : numpy.ndarray
        A 2d array. This is the block that gets repeated.
    N : int, Optional
        The number of copies.

    Returns
    -------
    numpy.ndarray
        A 2d numpy array.
    """
    nR, nC = a.shape[:2]
    res = np.zeros((nR * N, nC * N), dtype=a.dtype)
    for i in prange(N):
        _iR = i * nR
        iR_ = _iR + nR
        _iC = i * nC
        iC_ = _iC + nC
        res[_iR:iR_, _iC:iC_] = a
    return res


@njit(nogil=True, parallel=True, cache=__cache)
def _to_range_1d_(vals: ndarray, source: ndarray, target: ndarray):
    res = np.zeros_like(vals)
    s0, s1 = source
    t0, t1 = target
    b = (t1 - t0) / (s1 - s0)
    a = (t0 + t1) / 2 - b * (s0 + s1) / 2
    for i in prange(res.shape[0]):
        res[i] = a + b * vals[i]
    return res


def to_range_1d(
    vals: ndarray | Iterable,
    *_,
    source: ndarray,
    target: ndarray = None,
):
    if not isinstance(vals, ndarray):
        vals = np.array(
            [
                vals,
            ]
        )
    source = np.array([0.0, 1.0]) if source is None else np.array(source)
    target = np.array([-1.0, 1.0]) if target is None else np.array(target)
    return _to_range_1d_(vals, source, target)


# !FIXME : assumes a unique input array
@njit(nogil=True, parallel=True, fastmath=True, cache=__cache)
def find1d(arr, space):
    res = np.zeros(arr.shape, dtype=np.uint64)
    for i in prange(arr.shape[0]):
        res[i] = np.where(arr[i] == space)[0][0]
    return res


# !TODO generalize this up to n dimensions
# !FIXME : assumes a unique input array
@njit(nogil=True, parallel=True, fastmath=True, cache=__cache)
def find2d(arr, space):
    res = np.zeros(arr.shape, dtype=np.uint64)
    for i in prange(arr.shape[0]):
        res[i] = find1d(arr[i], space)
    return res



from .ls import moving_least_squares, least_squares, weighted_least_squares
from .functions import (
    CubicWeightFunction,
    ConstantWeightFunction,
    SingularWeightFunction,
)
from .lagrange import gen_Lagrange_1d, approx_Lagrange_1d
from .mls import MLSApproximator

__all__ = [
    "moving_least_squares",
    "least_squares",
    "weighted_least_squares",
    "CubicWeightFunction",
    "ConstantWeightFunction",
    "SingularWeightFunction",
    "gen_Lagrange_1d",
    "approx_Lagrange_1d",
    "MLSApproximator",
]
